"""

Revision ID: d663166c531d
Revises: fc191121f522
Create Date: 2025-07-09 17:44:25.728075

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "d663166c531d"
down_revision: Union[str, None] = "fc191121f522"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("median_weekly_growth", sa.Float(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("median_monthly_growth", sa.Float(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("median_yearly_growth", sa.Float(), nullable=True)
        )
        batch_op.create_index(
            "ix_analysis_median_monthly_growth", ["median_monthly_growth"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_median_weekly_growth", ["median_weekly_growth"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_median_yearly_growth", ["median_yearly_growth"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_median_yearly_growth")
        batch_op.drop_index("ix_analysis_median_weekly_growth")
        batch_op.drop_index("ix_analysis_median_monthly_growth")
        batch_op.drop_column("median_yearly_growth")
        batch_op.drop_column("median_monthly_growth")
        batch_op.drop_column("median_weekly_growth")
    # ### end Alembic commands ###



from datetime import date
from typing import Dict, Any, List, Optional

from sqlmodel import Field, SQLModel
from sqlalchemy import Column, JSON
from bullish.analysis.analysis import Analysis
from bullish.analysis.backtest import BacktestResult
from bullish.analysis.filter import FilteredResults
from bullish.analysis.indicators import SignalSeries
from bullish.analysis.industry_views import IndustryView
from bullish.analysis.openai import OpenAINews

from bullish.jobs.models import JobTracker
from sqlalchemy import Index


class BaseTable(SQLModel):
    symbol: str = Field(primary_key=True)
    source: str = Field(primary_key=True)


dynamic_indexes = tuple(
    Index(f"ix_analysis_{col}", col) for col in Analysis.model_fields
)


class OpenAINewsORM(SQLModel, OpenAINews, table=True):
    __tablename__ = "openai"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    symbol: str = Field(primary_key=True)
    news_date: date = Field(primary_key=True)


class AnalysisORM(BaseTable, Analysis, table=True):
    __tablename__ = "analysis"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    news_summary: Optional[List[Dict[str, Any]]] = Field(
        default=None, sa_column=Column(JSON)
    )
    summary: Optional[Dict[str, Any]] = Field(default=None, sa_column=Column(JSON))


AnalysisORM.__table_args__ = tuple(  # type: ignore # noqa: RUF005
    Index(f"ix_{AnalysisORM.__tablename__}_{col.name}", col)
    for col in AnalysisORM.__table__.columns
    if not col.primary_key and not col.index and col.name != "id"
) + (AnalysisORM.__table_args__,)


class JobTrackerORM(SQLModel, JobTracker, table=True):
    __tablename__ = "jobtracker"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    job_id: str = Field(primary_key=True)
    type: str  # type: ignore
    status: str  # type: ignore


class FilteredResultsORM(SQLModel, FilteredResults, table=True):
    __tablename__ = "filteredresults"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    name: str = Field(primary_key=True)
    symbols: list[str] = Field(sa_column=Column(JSON))
    filter_query: Dict[str, Any] = Field(sa_column=Column(JSON))  # type: ignore


class SignalSeriesORM(SQLModel, SignalSeries, table=True):
    __tablename__ = "signalseries"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    date: str = Field(primary_key=True)  # type: ignore
    name: str = Field(primary_key=True)
    symbol: str = Field(primary_key=True)
    value: float | None = Field(default=None, nullable=True)  # type: ignore


class IndustryViewORM(SQLModel, IndustryView, table=True):
    __tablename__ = "industryview"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    date: str = Field(primary_key=True)  # type: ignore
    created_at: str = Field(default=None, nullable=True)  # type: ignore
    simple_return: float | None = Field(default=None, nullable=True)  # type: ignore
    log_return: float | None = Field(default=None, nullable=True)  # type: ignore
    normalized_close: float | None = Field(default=None, nullable=True)  # type: ignore
    country: str = Field(primary_key=True)  # type: ignore
    industry: str = Field(primary_key=True)  # type: ignore
    industry_group: str | None = Field(default=None, nullable=True)  # type: ignore
    sector: str | None = Field(default=None, nullable=True)  # type: ignore
    type: str = Field(primary_key=True)  # type: ignore


class BacktestResultORM(SQLModel, BacktestResult, table=True):
    __tablename__ = "backtestresult"
    __table_args__ = {"extend_existing": True}  # noqa:RUF012
    name: str = Field(primary_key=True)
    start: str = Field(primary_key=True)  # type: ignore
    holding_period: int = Field(primary_key=True)
    extend_days: int = Field(primary_key=True)
    percentage: int = Field(primary_key=True)
    iterations: int = Field(primary_key=True)
    data: Dict[str, Any] = Field(sa_column=Column(JSON))



"""

Revision ID: 6d252e23f543
Revises: 12889a2cbd7d
Create Date: 2025-07-27 16:46:41.885125

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "6d252e23f543"
down_revision: Union[str, None] = "12889a2cbd7d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "backtestresult",
        sa.Column("end", sa.Date(), nullable=False),
        sa.Column("investment", sa.Float(), nullable=False),
        sa.Column("start", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("holding_period", sa.Integer(), nullable=False),
        sa.Column("extend_days", sa.Integer(), nullable=False),
        sa.Column("percentage", sa.Integer(), nullable=False),
        sa.Column("iterations", sa.Integer(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("data", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint(
            "start", "holding_period", "extend_days", "percentage", "iterations", "name"
        ),
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_table("backtestresult")
    # ### end Alembic commands ###



import logging
from datetime import date
from typing import Optional, Callable, cast

import numpy as np
import pandas as pd
import pandas_ta as ta

from pydantic import BaseModel

logger = logging.getLogger(__name__)
try:
    import talib
except Exception:
    logger.warning("Talib is not installed, skipping analysis")


def cross_simple(
    series_a: pd.Series, series_b: pd.Series, above: bool = True
) -> pd.Series:
    crossing = ta.cross(x=series_a, y=series_b, above=above)  # type: ignore
    return crossing  # type: ignore


def cross(
    series_a: pd.Series, series_b: pd.Series, above: bool = True
) -> Optional[date]:
    crossing = cross_simple(series_a=series_a, series_b=series_b, above=above)
    if not crossing[crossing == 1].index.empty:
        return crossing[crossing == 1].last_valid_index().date()  # type: ignore
    return None


def cross_value(series: pd.Series, number: int, above: bool = True) -> Optional[date]:
    return cross(series, pd.Series(number, index=series.index), above=above)


def cross_value_series(
    series_a: pd.Series, number: int, above: bool = True
) -> pd.Series:
    crossing = cross_simple(
        series_a, pd.Series(number, index=series_a.index), above=above
    )
    return crossing


def compute_adx(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["ADX_14"] = talib.ADX(data.high, data.low, close=data.close)  # type: ignore
    results["MINUS_DI"] = talib.MINUS_DI(data.high, data.low, data.close)  # type: ignore
    results["PLUS_DI"] = talib.PLUS_DI(data.high, data.low, data.close)  # type: ignore
    return results


def compute_pandas_ta_adx(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    adx = ta.adx(data.high, data.low, data.close, length=14)
    results["ADX_14"] = adx["ADX_14"]
    results["MINUS_DI"] = adx["DMN_14"]
    results["PLUS_DI"] = adx["DMP_14"]
    return results


def compute_macd(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    (
        results["MACD_12_26_9"],
        results["MACD_12_26_9_SIGNAL"],
        results["MACD_12_26_9_HIST"],
    ) = talib.MACD(
        data.close  # type: ignore
    )
    return results


def compute_pandas_ta_macd(data: pd.DataFrame) -> pd.DataFrame:

    macd = ta.macd(data.close, fast=12, slow=26, signal=9)
    results = pd.DataFrame(index=macd.index)
    results["MACD_12_26_9"] = macd["MACD_12_26_9"]
    results["MACD_12_26_9_SIGNAL"] = macd["MACDs_12_26_9"]
    results["MACD_12_26_9_HIST"] = macd["MACDh_12_26_9"]
    return results


def compute_rsi(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["RSI"] = talib.RSI(data.close)  # type: ignore
    results["CLOSE"] = data.close
    return results


def compute_pandas_ta_rsi(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["RSI"] = ta.rsi(data.close, length=14)
    results["CLOSE"] = data.close
    return results


def compute_stoch(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["SLOW_K"], results["SLOW_D"] = talib.STOCH(data.high, data.low, data.close)  # type: ignore
    return results


def compute_pandas_ta_stoch(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    stoch = ta.stoch(data.high, data.low, data.close, k=5, d=3, smooth_k=3)
    results["SLOW_K"] = stoch["STOCHk_5_3_3"]
    results["SLOW_D"] = stoch["STOCHd_5_3_3"]
    return results


def compute_mfi(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["MFI"] = talib.MFI(data.high, data.low, data.close, data.volume)  # type: ignore
    return results


def compute_pandas_ta_mfi(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["MFI"] = ta.mfi(data.high, data.low, data.close, data.volume, length=14)
    return results


def compute_roc(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["ROC_7"] = talib.ROC(data.close, timeperiod=7)  # type: ignore
    results["ROC_1"] = talib.ROC(data.close, timeperiod=1)  # type: ignore
    results["ROC_30"] = talib.ROC(data.close, timeperiod=30)  # type: ignore
    mom = talib.MOM(data.close, timeperiod=252)  # type: ignore
    results["MOM"] = mom.shift(21)  # type: ignore

    return results


def compute_pandas_ta_roc(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["ROC_7"] = ta.roc(data.close, length=7)
    results["ROC_1"] = ta.roc(data.close, length=1)
    results["ROC_30"] = ta.roc(data.close, length=30)
    return results


def compute_sma(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["SMA_50"] = talib.SMA(data.close, timeperiod=50)  # type: ignore
    results["SMA_200"] = talib.SMA(data.close, timeperiod=200)  # type: ignore
    results["CLOSE"] = data.close
    return results


def compute_pandas_ta_sma(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["SMA_50"] = ta.sma(data.close, length=50)
    results["SMA_200"] = ta.sma(data.close, length=200)
    results["CLOSE"] = data.close

    return results


def compute_adosc(data: pd.DataFrame) -> pd.DataFrame:
    data_ = data.copy()
    results = pd.DataFrame(index=data.index)
    results["ADOSC"] = talib.ADOSC(data.high, data.low, data.close, data.volume)  # type: ignore
    data_["ADOSC"] = results["ADOSC"]
    data_["HIGHEST_20"] = data_.close.rolling(window=20).max()
    results["ADOSC_SIGNAL"] = (data_.close > data_["HIGHEST_20"].shift(1)) & (
        data_["ADOSC"] > 0
    )
    return results


def compute_pandas_ta_adosc(data: pd.DataFrame) -> pd.DataFrame:
    data_ = data.copy()
    results = pd.DataFrame(index=data.index)
    results["ADOSC"] = ta.adosc(data.high, data.low, data.close, data.volume)
    data_["ADOSC"] = results["ADOSC"]
    data_["HIGHEST_20"] = data_.close.rolling(window=20).max()
    results["ADOSC_SIGNAL"] = (data_.close > data_["HIGHEST_20"].shift(1)) & (
        data_["ADOSC"] > 0
    )
    return results


def compute_ad(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["AD"] = talib.AD(data.high, data.low, data.close, data.volume)  # type: ignore
    return results


def compute_pandas_ta_ad(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["AD"] = ta.ad(data.high, data.low, data.close, data.volume)
    return results


def compute_obv(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["OBV"] = talib.OBV(data.close, data.volume)  # type: ignore
    return results


def compute_pandas_ta_obv(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["OBV"] = ta.obv(data.close, data.volume)
    return results


def compute_atr(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["ATR"] = talib.ATR(data.high, data.low, data.close)  # type: ignore
    return results


def compute_pandas_ta_atr(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["ATR"] = ta.atr(data.high, data.low, data.close, length=14)
    return results


def compute_natr(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["NATR"] = talib.NATR(data.high, data.low, data.close)  # type: ignore
    return results


def compute_pandas_ta_natr(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["NATR"] = ta.natr(data.high, data.low, data.close, length=14)
    return results


def compute_trange(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["TRANGE"] = talib.TRANGE(data.high, data.low, data.close)  # type: ignore
    return results


def compute_pandas_ta_trange(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["TRANGE"] = ta.true_range(data.high, data.low, data.close)
    return results


def compute_patterns(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["CDLMORNINGSTAR"] = talib.CDLMORNINGSTAR(
        data.open, data.high, data.low, data.close  # type: ignore
    )
    results["CDL3LINESTRIKE"] = talib.CDL3LINESTRIKE(
        data.open, data.high, data.low, data.close  # type: ignore
    )
    results["CDL3WHITESOLDIERS"] = talib.CDL3WHITESOLDIERS(
        data.open, data.high, data.low, data.close  # type: ignore
    )
    results["CDLABANDONEDBABY"] = talib.CDLABANDONEDBABY(
        data.open, data.high, data.low, data.close  # type: ignore
    )
    results["CDLTASUKIGAP"] = talib.CDLTASUKIGAP(
        data.open, data.high, data.low, data.close  # type: ignore
    )
    results["CDLPIERCING"] = talib.CDLPIERCING(
        data.open, data.high, data.low, data.close  # type: ignore
    )
    results["CDLENGULFING"] = talib.CDLENGULFING(
        data.open, data.high, data.low, data.close  # type: ignore
    )
    return results


def perc(data: pd.Series) -> float:
    if len(data) < 2 or data.iloc[0] == 0:
        return np.nan
    return cast(float, ((data.iloc[-1] - data.iloc[0]) / data.iloc[0]) * 100)


def compute_price(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["200_DAY_HIGH"] = data.close.rolling(window=200).max()
    results["200_DAY_LOW"] = data.close.rolling(window=200).min()
    results["20_DAY_HIGH"] = data.close.rolling(window=20).max()
    results["20_DAY_LOW"] = data.close.rolling(window=20).min()
    results["LAST_PRICE"] = data.close
    results["WEEKLY_GROWTH"] = data.close.resample("W").transform(perc).ffill()  # type: ignore
    results["MONTHLY_GROWTH"] = data.close.resample("ME").transform(perc).ffill()  # type: ignore
    results["YEARLY_GROWTH"] = data.close.resample("YE").transform(perc).ffill()  # type: ignore
    return results


def compute_volume(data: pd.DataFrame) -> pd.DataFrame:
    results = pd.DataFrame(index=data.index)
    results["AVERAGE_VOLUME_10"] = data.volume.rolling(window=10).mean()
    results["AVERAGE_VOLUME_30"] = data.volume.rolling(window=30).mean()
    results["VOLUME"] = data.volume
    return results


def find_last_true_run_start(series: pd.Series) -> Optional[date]:
    if not series.iloc[-1]:
        return None
    arr = series.to_numpy()
    change_points = np.flatnonzero(np.r_[True, arr[1:] != arr[:-1]])
    run_starts = change_points
    true_runs = run_starts[arr[run_starts]]
    last_true_run_start = true_runs[-1]
    return series.index[last_true_run_start].date()  # type: ignore


def sma_50_above_sma_200(data: pd.DataFrame) -> Optional[date]:
    date_1 = find_last_true_run_start(data.SMA_50 > data.SMA_200)
    return date_1


def price_above_sma50(data: pd.DataFrame) -> Optional[date]:
    date_1 = find_last_true_run_start(data.SMA_50 < data.CLOSE)
    return date_1


class IndicatorFunction(BaseModel):
    expected_columns: list[str]
    functions: list[Callable[[pd.DataFrame], pd.DataFrame]]

    def call(self, data: pd.DataFrame) -> pd.DataFrame:
        data_ = None
        for function in self.functions:
            try:
                data_ = function(data)
                break
            except Exception as e:
                logger.error(f"Fail to compute function {function.__name__}: {e}")
        if data_ is None:
            raise ValueError(
                f"No data returned from indicator functions with expected columns {self.expected_columns}."
            )
        if not set(self.expected_columns).issubset(set(data_.columns)):
            raise ValueError(
                f"Expected columns {self.expected_columns} not found in data columns {data_.columns.tolist()}"
            )
        return data_


ADX = IndicatorFunction(
    expected_columns=["ADX_14", "MINUS_DI", "PLUS_DI"],
    functions=[compute_adx, compute_pandas_ta_adx],
)
MACD = IndicatorFunction(
    expected_columns=["MACD_12_26_9", "MACD_12_26_9_SIGNAL", "MACD_12_26_9_HIST"],
    functions=[compute_macd, compute_pandas_ta_macd],
)
RSI = IndicatorFunction(
    expected_columns=["RSI"], functions=[compute_rsi, compute_pandas_ta_rsi]
)
STOCH = IndicatorFunction(
    expected_columns=["SLOW_K", "SLOW_D"],
    functions=[compute_stoch, compute_pandas_ta_stoch],
)
MFI = IndicatorFunction(
    expected_columns=["MFI"], functions=[compute_mfi, compute_pandas_ta_mfi]
)
ROC = IndicatorFunction(
    expected_columns=["ROC_7", "ROC_1", "ROC_30", "MOM"],
    functions=[compute_roc, compute_pandas_ta_roc],
)
CANDLESTOCK_PATTERNS = IndicatorFunction(
    expected_columns=[
        "CDLMORNINGSTAR",
        "CDL3LINESTRIKE",
        "CDL3WHITESOLDIERS",
        "CDLABANDONEDBABY",
        "CDLTASUKIGAP",
        "CDLPIERCING",
        "CDLENGULFING",
    ],
    functions=[compute_patterns],
)

SMA = IndicatorFunction(
    expected_columns=["SMA_50", "SMA_200", "CLOSE"],
    functions=[compute_sma, compute_pandas_ta_sma],
)

ADOSC = IndicatorFunction(
    expected_columns=["ADOSC", "ADOSC_SIGNAL"],
    functions=[compute_adosc, compute_pandas_ta_adosc],
)

AD = IndicatorFunction(
    expected_columns=["AD"],
    functions=[compute_ad, compute_pandas_ta_ad],
)
OBV = IndicatorFunction(
    expected_columns=["OBV"],
    functions=[compute_obv, compute_pandas_ta_obv],
)
ATR = IndicatorFunction(
    expected_columns=["ATR"],
    functions=[compute_atr, compute_pandas_ta_atr],
)
NATR = IndicatorFunction(
    expected_columns=["NATR"],
    functions=[compute_natr, compute_pandas_ta_natr],
)
TRANGE = IndicatorFunction(
    expected_columns=["TRANGE"],
    functions=[compute_trange, compute_pandas_ta_trange],
)
VOLUME = IndicatorFunction(
    expected_columns=["AVERAGE_VOLUME_10", "AVERAGE_VOLUME_30", "VOLUME"],
    functions=[compute_volume],
)
PRICE = IndicatorFunction(
    expected_columns=[
        "200_DAY_HIGH",
        "200_DAY_LOW",
        "20_DAY_HIGH",
        "20_DAY_LOW",
        "LAST_PRICE",
        "WEEKLY_GROWTH",
        "MONTHLY_GROWTH",
        "YEARLY_GROWTH",
    ],
    functions=[compute_price],
)


def add_indicators(data: pd.DataFrame) -> pd.DataFrame:
    indicators = [ADX, MACD, RSI, STOCH, SMA, ADOSC, AD, OBV, ATR, NATR, TRANGE]
    expected_columns = [c for i in indicators for c in i.expected_columns]
    for indicator in indicators:
        data = pd.concat([data, indicator.call(data)], axis=1)
    if not set(expected_columns).issubset(set(data.columns)):
        raise ValueError(
            f"Expected columns {expected_columns} not found in data columns {data.columns.tolist()}"
        )
    return data


class Line(BaseModel):
    value: float
    previous: float


class SupportResistance(BaseModel):
    support: Line
    resistance: Line


def support_resistance(df: pd.DataFrame, window: int = 5) -> SupportResistance:

    w = window * 2 + 1
    highs = df.high.rolling(w, center=True).max()
    lows = df.low.rolling(w, center=True).min()
    swing_high_mask = df.high == highs
    swing_low_mask = df.low == lows

    raw_res = df.loc[swing_high_mask, "high"].to_numpy()
    raw_sup = df.loc[swing_low_mask, "low"].to_numpy()
    return SupportResistance(
        support=Line(value=float(raw_sup[-1]), previous=float(raw_sup[-2])),
        resistance=Line(value=float(raw_res[-1]), previous=float(raw_res[-2])),
    )


def bollinger_bands(
    data: pd.DataFrame, window: int = 20, std_dev: float = 2.0
) -> pd.DataFrame:
    bbands = ta.bbands(
        data.close, timeperiod=window, nbdevup=std_dev, nbdevdn=std_dev, matype=0  # type: ignore
    )
    return bbands



"""

Revision ID: 11d35a452b40
Revises: 73564b60fe24
Create Date: 2025-06-23 06:02:07.122505

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision: str = "11d35a452b40"
down_revision: Union[str, None] = "73564b60fe24"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.alter_column(
            "quarterly_positive_free_cash_flow",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_operating_cash_flow",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_operating_cash_flow_is_higher_than_net_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_net_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_operating_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_net_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_operating_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_diluted_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_basic_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_basic_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_diluted_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_debt_to_equity",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_return_on_assets",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_return_on_equity",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_free_cash_flow",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_operating_cash_flow",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "operating_cash_flow_is_higher_than_net_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_net_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_operating_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_net_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_operating_income",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_diluted_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_basic_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_basic_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_diluted_eps",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_debt_to_equity",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_return_on_assets",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_return_on_equity",
            existing_type=sa.FLOAT(),
            type_=sa.Boolean(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.alter_column(
            "positive_return_on_equity",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_return_on_assets",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_debt_to_equity",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_diluted_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_basic_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_basic_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_diluted_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_operating_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_net_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_operating_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_net_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "operating_cash_flow_is_higher_than_net_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "growing_operating_cash_flow",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "positive_free_cash_flow",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_return_on_equity",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_return_on_assets",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_debt_to_equity",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_diluted_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_basic_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_basic_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_diluted_eps",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_operating_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_net_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_operating_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_net_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_operating_cash_flow_is_higher_than_net_income",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_growing_operating_cash_flow",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "quarterly_positive_free_cash_flow",
            existing_type=sa.Boolean(),
            type_=sa.FLOAT(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



"""

Revision ID: ff0cc4ba40ec
Revises: 79bc71ec6f9e
Create Date: 2025-08-05 12:09:12.108606

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "ff0cc4ba40ec"
down_revision: Union[str, None] = "79bc71ec6f9e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("average_volume_10", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("average_volume_30", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("volume_above_average", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("weekly_growth", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("monthly_growth", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("yearly_growth", sa.Float(), nullable=True))
        batch_op.create_index(
            "ix_analysis_average_volume_10", ["average_volume_10"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_average_volume_30", ["average_volume_30"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_monthly_growth", ["monthly_growth"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_volume_above_average", ["volume_above_average"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_weekly_growth", ["weekly_growth"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_yearly_growth", ["yearly_growth"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_yearly_growth")
        batch_op.drop_index("ix_analysis_weekly_growth")
        batch_op.drop_index("ix_analysis_volume_above_average")
        batch_op.drop_index("ix_analysis_monthly_growth")
        batch_op.drop_index("ix_analysis_average_volume_30")
        batch_op.drop_index("ix_analysis_average_volume_10")
        batch_op.drop_column("yearly_growth")
        batch_op.drop_column("monthly_growth")
        batch_op.drop_column("weekly_growth")
        batch_op.drop_column("volume_above_average")
        batch_op.drop_column("average_volume_30")
        batch_op.drop_column("average_volume_10")

    # ### end Alembic commands ###



from __future__ import annotations

import os
import subprocess
import signal
import sys
import time
import atexit
from pathlib import Path
from typing import Any

import typer
from dotenv import load_dotenv

load_dotenv(dotenv_path=Path.cwd().joinpath(".env"))

app = typer.Typer(add_completion=False)
STREAMLIT_FILE = Path(__file__).parent.joinpath("app", "app.py")


@app.command()
def serve(  # noqa: C901
    host: str = typer.Option("0.0.0.0", help="Streamlit host"),  # noqa: S104
    port: int = typer.Option(8501, help="Streamlit port"),
    env_vars: Path = typer.Option(  # noqa: B008
        None, help="Environment variables file"
    ),
    custom_filters: Path = typer.Option(  # noqa: B008
        None, help="Environment variables file"
    ),
) -> None:
    if env_vars:
        env_vars_path = Path(env_vars)
        if env_vars_path.exists():
            load_dotenv(dotenv_path=env_vars_path)
    if custom_filters:
        custom_filters_path = Path(custom_filters)
        if custom_filters_path.exists():
            os.environ["CUSTOM_FILTERS_PATH"] = str(custom_filters_path)
    children: list[subprocess.Popen] = []  # type: ignore

    def _shutdown(*_: Any) -> None:
        for p in children:
            if p.poll() is None:
                p.send_signal(signal.SIGINT)
                try:
                    p.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    p.kill()
        sys.exit(0)

    signal.signal(signal.SIGINT, _shutdown)
    signal.signal(signal.SIGTERM, _shutdown)
    atexit.register(_shutdown)

    try:
        children.append(
            subprocess.Popen(
                [  # noqa: S603
                    sys.executable,
                    "-m",
                    "huey.bin.huey_consumer",
                    "bullish.jobs.tasks.huey",
                    "--workers",
                    "2",
                ],
                env=os.environ.copy(),
            )
        )
        children.append(
            subprocess.Popen(
                [  # noqa: S603
                    sys.executable,
                    "-m",
                    "streamlit",
                    "run",
                    str(STREAMLIT_FILE),
                    "--server.address",
                    host,
                    "--server.port",
                    str(port),
                    os.devnull,
                ],
                env=os.environ.copy(),
            )
        )
        while True:
            time.sleep(1)

    except Exception as exc:  # pragma: no cover
        typer.secho(
            f"❌ Failed to start services: {exc}", fg=typer.colors.RED, err=True
        )
        _shutdown()


if __name__ == "__main__":
    app()



import datetime
import json
import os
from datetime import timedelta
from pathlib import Path
from typing import Dict, Any, Optional, List, Union, get_args, Tuple

from bullish.analysis.analysis import AnalysisView
from bullish.analysis.backtest import (
    BacktestQueryDate,
    BacktestQueries,
    BacktestQueryRange,
    BacktestQuerySelection,
)
from bullish.analysis.constants import (
    Europe,
    Us,
    HighGrowthIndustry,
    DefensiveIndustries,
)
from bullish.analysis.filter import FilterQuery, BOOLEAN_GROUP_MAPPING
from pydantic import BaseModel, Field

from bullish.analysis.indicators import Indicators
from bullish.database.crud import BullishDb

DATE_THRESHOLD = [
    datetime.date.today() - datetime.timedelta(days=2),
    datetime.date.today(),
]


def _get_variants(variants: List[str]) -> List[Tuple[str, ...]]:
    return [tuple(variants[:i]) for i in range(1, len(variants) + 1)]


class NamedFilterQuery(FilterQuery):
    name: str
    description: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump(
            exclude_unset=True,
            exclude_none=True,
            exclude_defaults=True,
            exclude={"name"},
        )

    def to_backtesting_query(
        self, backtest_start_date: datetime.date
    ) -> BacktestQueries:
        queries: List[
            Union[BacktestQueryRange, BacktestQueryDate, BacktestQuerySelection]
        ] = []
        in_use_backtests = Indicators().in_use_backtest()
        for in_use in in_use_backtests:
            value = self.to_dict().get(in_use)
            if value and self.model_fields[in_use].annotation == List[datetime.date]:
                delta = value[1] - value[0]
                queries.append(
                    BacktestQueryDate(
                        name=in_use.upper(),
                        start=backtest_start_date - delta,
                        end=backtest_start_date,
                        table="signalseries",
                    )
                )
        for field in self.to_dict():
            if field in BOOLEAN_GROUP_MAPPING:
                value = self.to_dict().get(field)
                if value and self.model_fields[field].annotation == Optional[List[str]]:  # type: ignore
                    queries.extend(
                        [
                            BacktestQueryDate(
                                name=v.upper(),
                                start=backtest_start_date - timedelta(days=252),
                                end=backtest_start_date,
                                table="signalseries",
                            )
                            for v in value
                        ]
                    )

            if field in AnalysisView.model_fields:
                value = self.to_dict().get(field)
                if (
                    value
                    and self.model_fields[field].annotation == Optional[List[float]]  # type: ignore
                    and len(value) == 2
                ):
                    queries.append(
                        BacktestQueryRange(
                            name=field.lower(),
                            min=value[0],
                            max=value[1],
                            table="analysis",
                        )
                    )
                if value and self.model_fields[field].annotation == Optional[List[str]]:  # type: ignore
                    queries.append(
                        BacktestQuerySelection(
                            name=field.lower(),
                            selections=value,
                            table="analysis",
                        )
                    )

        return BacktestQueries(queries=queries)

    def get_backtesting_symbols(
        self, bullish_db: BullishDb, backtest_start_date: datetime.date
    ) -> List[str]:
        queries = self.to_backtesting_query(backtest_start_date)

        return bullish_db.read_query(queries.to_query())["symbol"].tolist()  # type: ignore

    def country_variant(self, suffix: str, countries: List[str]) -> "NamedFilterQuery":
        return NamedFilterQuery.model_validate(
            self.model_dump()
            | {"name": f"{self.name} ({suffix})", "country": countries}
        )

    def update_indicator_filter(
        self, suffix: str, rsi_parameter_name: str
    ) -> "NamedFilterQuery":
        return NamedFilterQuery.model_validate(
            self.model_dump()
            | {"name": f"{self.name} ({suffix})", rsi_parameter_name: DATE_THRESHOLD}
        )

    def _custom_variant(
        self, suffix: str, properties: Dict[str, Any]
    ) -> "NamedFilterQuery":
        return NamedFilterQuery.model_validate(
            self.model_dump() | {"name": f"{self.name} ({suffix})", **properties}
        )

    def week_top_performers(self) -> "NamedFilterQuery":
        properties = {
            "volume_above_average": DATE_THRESHOLD,
            "weekly_growth": [1, 100],
        }
        return self._custom_variant("Week Top Performers", properties)

    def month_top_performers(self) -> "NamedFilterQuery":
        properties = {
            "monthly_growth": [8, 100],
        }
        return self._custom_variant("Month Top Performers", properties)

    def year_top_performers(self) -> "NamedFilterQuery":
        properties = {
            "volume_above_average": DATE_THRESHOLD,
            "sma_50_above_sma_200": [
                datetime.date.today() - datetime.timedelta(days=5000),
                datetime.date.today(),
            ],
            "yearly_growth": [30, 100],
        }
        return self._custom_variant("Yearly Top Performers", properties)

    def poor_performers(self) -> "NamedFilterQuery":
        properties = {
            "sma_50_below_sma_200": [
                datetime.date.today() - datetime.timedelta(days=5000),
                datetime.date.today(),
            ],
            "price_below_sma_50": [
                datetime.date.today() - datetime.timedelta(days=5000),
                datetime.date.today(),
            ],
            "monthly_growth": [-100, 0],
        }
        return self._custom_variant("Poor Performers", properties)

    def yearly_fundamentals(self) -> "NamedFilterQuery":
        properties = {
            "income": [
                "positive_operating_income",
                "positive_net_income",
                "growing_net_income",
                "growing_operating_income",
            ],
            "cash_flow": ["positive_free_cash_flow", "growing_operating_cash_flow"],
            "properties": [
                "positive_return_on_equity",
                "operating_cash_flow_is_higher_than_net_income",
            ],
        }
        return self._custom_variant("Yearly Fundamentals", properties)

    def quarterly_fundamentals(self) -> "NamedFilterQuery":
        properties = {
            "income": [
                "quarterly_positive_operating_income",
                "quarterly_positive_net_income",
            ],
            "cash_flow": [
                "quarterly_positive_free_cash_flow",
            ],
            "properties": [
                "quarterly_operating_cash_flow_is_higher_than_net_income",
            ],
        }
        return self._custom_variant("Quarterly Fundamentals", properties)

    def growing_quarterly_fundamentals(self) -> "NamedFilterQuery":
        properties = {
            "income": [
                "quarterly_positive_operating_income",
                "quarterly_positive_net_income",
                "quarterly_growing_net_income",
            ],
            "cash_flow": [
                "quarterly_positive_free_cash_flow",
                "quarterly_growing_operating_cash_flow",
            ],
            "properties": [
                "quarterly_operating_cash_flow_is_higher_than_net_income",
            ],
        }
        return self._custom_variant("Growing Quarterly Fundamentals", properties)

    def min_fundamentals(self) -> "NamedFilterQuery":
        properties = {
            "income": [
                "positive_operating_income",
                "positive_net_income",
            ],
            "cash_flow": [
                "positive_free_cash_flow",
            ],
            "eps": [
                "positive_diluted_eps",  # or positive_basic_eps if diluted not available
            ],
            "properties": [
                "positive_return_on_equity",
                "operating_cash_flow_is_higher_than_net_income",
            ],
        }
        return self._custom_variant("Min Fundamentals", properties)

    def high_growth(self) -> "NamedFilterQuery":
        properties = {"industry": list(get_args(HighGrowthIndustry))}
        return self._custom_variant("Growth", properties)

    def defensive(self) -> "NamedFilterQuery":
        properties = {"industry": list(get_args(DefensiveIndustries))}
        return self._custom_variant("Defensive", properties)

    def cheap(self) -> "NamedFilterQuery":
        properties = {"last_price": [1, 30]}
        return self._custom_variant("Cheap", properties)

    def europe(self) -> "NamedFilterQuery":
        return self.country_variant("Europe", list(get_args(Europe)))

    def us(self) -> "NamedFilterQuery":
        return self.country_variant("Us", list(get_args(Us)))

    def rsi_30(self) -> "NamedFilterQuery":
        return self.update_indicator_filter("RSI 30", "rsi_bullish_crossover_30")

    def rsi_40(self) -> "NamedFilterQuery":
        return self.update_indicator_filter("RSI 40", "rsi_bullish_crossover_40")

    def macd(self) -> "NamedFilterQuery":
        return self.update_indicator_filter("MACD", "macd_12_26_9_bullish_crossover")

    def rsi_neutral_(self) -> "NamedFilterQuery":
        return self.update_indicator_filter("RSI Neutral", "rsi_neutral")

    def rsi_oversold_(self) -> "NamedFilterQuery":
        return self.update_indicator_filter("RSI Oversold", "rsi_oversold")

    def rsi_overbought_(self) -> "NamedFilterQuery":
        return self.update_indicator_filter("RSI Overbought", "rsi_overbought")

    def adx(self) -> "NamedFilterQuery":
        return self.update_indicator_filter("ADX 14", "adx_14")

    def earnings_date(self) -> "NamedFilterQuery":
        return NamedFilterQuery.model_validate(
            self.model_dump()
            | {
                "name": f"{self.name} (Earnings Date)",
                "next_earnings_date": [
                    datetime.date.today(),
                    datetime.date.today() + timedelta(days=20),
                ],
            }
        )

    def variants(
        self,
        variants: Optional[List[List[str]]] = None,
        filters: Optional[List[str]] = None,
    ) -> List["NamedFilterQuery"]:
        if filters and self.name not in filters:
            return [self]
        variants = variants or [["europe"], ["us"]]

        _variants = {v for variant in variants for v in _get_variants(variant)}
        filters_ = []
        for attributes in _variants:
            filter__ = self
            for attr in attributes:
                filter__ = getattr(filter__, attr)()
            filters_.append(filter__)

        return [self, *filters_]


def load_custom_filters() -> List[NamedFilterQuery]:
    if "CUSTOM_FILTERS_PATH" in os.environ:
        custom_filters_path = os.environ["CUSTOM_FILTERS_PATH"]
        return [
            variant
            for f in read_custom_filters(Path(custom_filters_path))
            for variant in f.variants(
                variants=[["rsi_overbought_"]], filters=["portfolio", "Portfolio"]
            )
        ]
    return []


def read_custom_filters(custom_filters_path: Path) -> List[NamedFilterQuery]:
    if custom_filters_path.exists():
        filters = json.loads(custom_filters_path.read_text())
        return [NamedFilterQuery.model_validate(filter) for filter in filters]
    return []


SMALL_CAP = NamedFilterQuery(
    name="Small Cap",
    last_price=[1, 20],
    market_capitalization=[5e7, 5e8],
    properties=["positive_debt_to_equity"],
    average_volume_30=[50000, 5e9],
    order_by_desc="market_capitalization",
).variants(
    variants=[
        ["week_top_performers", "min_fundamentals"],
        ["month_top_performers", "min_fundamentals"],
        ["earnings_date", "min_fundamentals"],
        ["rsi_oversold_", "min_fundamentals"],
    ]
)

LARGE_CAPS = NamedFilterQuery(
    name="Large Cap",
    order_by_desc="market_capitalization",
    market_capitalization=[1e10, 1e14],
).variants(
    variants=[
        ["rsi_oversold_", "macd", "yearly_fundamentals"],
        ["rsi_neutral_", "macd", "adx", "yearly_fundamentals"],
        ["rsi_30", "macd", "adx", "yearly_fundamentals"],
        ["rsi_oversold_", "macd", "quarterly_fundamentals"],
        ["rsi_neutral_", "macd", "adx", "quarterly_fundamentals"],
        ["rsi_30", "macd", "adx", "quarterly_fundamentals"],
        ["earnings_date", "quarterly_fundamentals", "yearly_fundamentals"],
    ]
)

MID_CAPS = NamedFilterQuery(
    name="Mid Cap",
    order_by_desc="market_capitalization",
    market_capitalization=[5e8, 1e10],
).variants(
    variants=[
        ["week_top_performers"],
        ["month_top_performers"],
        ["earnings_date", "quarterly_fundamentals", "yearly_fundamentals"],
        ["rsi_oversold_", "macd", "adx"],
    ]
)


def predefined_filters() -> list[NamedFilterQuery]:
    return [
        *load_custom_filters(),
        *SMALL_CAP,
        *MID_CAPS,
        *LARGE_CAPS,
    ]


class PredefinedFilters(BaseModel):
    filters: list[NamedFilterQuery] = Field(default_factory=predefined_filters)

    def get_predefined_filter_names(self) -> list[str]:
        return [filter.name for filter in self.filters]

    def get_predefined_filter(self, name: str) -> Dict[str, Any]:
        for filter in self.filters:
            if filter.name == name:
                return filter.to_dict()
        raise ValueError(f"Filter with name '{name}' not found.")



"""

Revision ID: 260fcff7212e
Revises: 4ee82b171449
Create Date: 2025-08-11 10:08:17.582390

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "260fcff7212e"
down_revision: Union[str, None] = "4ee82b171449"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("sma_50_below_sma_200", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("price_below_sma_50", sa.Date(), nullable=True))
        batch_op.create_index(
            "ix_analysis_price_below_sma_50", ["price_below_sma_50"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_sma_50_below_sma_200", ["sma_50_below_sma_200"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_sma_50_below_sma_200")
        batch_op.drop_index("ix_analysis_price_below_sma_50")
        batch_op.drop_column("price_below_sma_50")
        batch_op.drop_column("sma_50_below_sma_200")

    # ### end Alembic commands ###



"""

Revision ID: b76079e9845f
Revises: bf6b86dd5463
Create Date: 2025-07-12 21:32:08.865721

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "b76079e9845f"
down_revision: Union[str, None] = "bf6b86dd5463"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("momentum_time_span", sa.Date(), nullable=True))
        batch_op.create_index(
            "ix_analysis_momentum_time_span", ["momentum_time_span"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_momentum_time_span")
        batch_op.drop_column("momentum_time_span")

    # ### end Alembic commands ###



"""

Revision ID: ee5baabb35f8
Revises: 08ac1116e055
Create Date: 2025-07-02 05:16:42.862366

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "ee5baabb35f8"
down_revision: Union[str, None] = "08ac1116e055"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("adosc_crosses_above_0", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("positive_adosc_20_day_breakout", sa.Date(), nullable=True)
        )
        batch_op.create_index(
            "ix_analysis_adosc_crosses_above_0", ["adosc_crosses_above_0"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_positive_adosc_20_day_breakout",
            ["positive_adosc_20_day_breakout"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_positive_adosc_20_day_breakout")
        batch_op.drop_index("ix_analysis_adosc_crosses_above_0")
        batch_op.drop_column("positive_adosc_20_day_breakout")
        batch_op.drop_column("adosc_crosses_above_0")

    # ### end Alembic commands ###



import json
import logging
from datetime import date
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Dict

import pandas as pd
from bearish.database.crud import BearishDb  # type: ignore
from bearish.database.schemas import EarningsDateORM, EquityORM, PriceORM  # type: ignore
from bearish.models.base import Ticker  # type: ignore
from bearish.models.price.price import Price  # type: ignore
from bearish.models.price.prices import Prices  # type: ignore
from bearish.types import Sources  # type: ignore
from pydantic import ConfigDict
from sqlalchemy import Engine, create_engine, insert, delete, update, inspect
from sqlalchemy import text
from sqlmodel import Session, select

from bullish.analysis.analysis import Analysis, SubjectAnalysis
from bullish.analysis.constants import Industry, IndustryGroup, Sector, Country
from bullish.analysis.filter import FilteredResults
from bullish.analysis.indicators import SignalSeries
from bullish.analysis.industry_views import Type, IndustryView

from bullish.database.schemas import (
    AnalysisORM,
    JobTrackerORM,
    FilteredResultsORM,
    IndustryViewORM,
    SignalSeriesORM,
    BacktestResultORM,
    OpenAINewsORM,
)
from bullish.database.scripts.upgrade import upgrade
from bullish.exceptions import DatabaseFileNotFoundError
from bullish.interface.interface import BullishDbBase
from bullish.jobs.models import JobTracker, JobTrackerStatus
from tickermood.database.scripts.upgrade import upgrade as tickermood_upgrade  # type: ignore

if TYPE_CHECKING:
    from bullish.analysis.backtest import BacktestResult, BacktestResultQuery
    from bullish.analysis.openai import OpenAINews

logger = logging.getLogger(__name__)

BATCH_SIZE = 5000


class BullishDb(BearishDb, BullishDbBase):  # type: ignore
    model_config = ConfigDict(arbitrary_types_allowed=True)
    database_path: Path

    def valid(self) -> bool:
        """Check if the database is valid."""
        return self.database_path.exists() and self.database_path.is_file()

    @cached_property
    def _engine(self) -> Engine:
        if not self.valid():
            raise DatabaseFileNotFoundError("Database file not found.")
        database_url = f"sqlite:///{Path(self.database_path)}"
        try:
            upgrade(self.database_path)
        except Exception as e:
            logger.warning(
                f"Failed to upgrade the database at {self.database_path}. "
                f"Reason: {e}"
                "Skipping upgrade. "
            )
        engine = create_engine(database_url)
        inspector = inspect(engine)
        if "subject" not in inspector.get_table_names():
            logger.info(
                "Running tickermood upgrade to create the subject table in the database."
            )
            try:
                tickermood_upgrade(database_url=database_url, no_migration=True)
            except Exception as e:
                logger.error(f"failed to update database: {e}")
                print(f"failed to update database: {e}")
        return engine

    def model_post_init(self, __context: Any) -> None:
        self._engine  # noqa: B018

    def _write_analysis(self, analysis: Analysis) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(AnalysisORM)
                .prefix_with("OR REPLACE")
                .values(analysis.model_dump())
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def _write_many_analysis(self, many_analysis: List[Analysis]) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(AnalysisORM)
                .prefix_with("OR REPLACE")
                .values([a.model_dump() for a in many_analysis])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def _read_analysis(self, ticker: Ticker) -> Optional[Analysis]:
        with Session(self._engine) as session:
            query = select(AnalysisORM).where(AnalysisORM.symbol == ticker.symbol)
            analysis = session.exec(query).first()
            if not analysis:
                return None
            return Analysis.model_validate(analysis)

    def read_symbols(self) -> List[str]:
        query = "SELECT DISTINCT symbol FROM analysis"
        data = pd.read_sql_query(query, self._engine)
        return data["symbol"].tolist()

    def _read_analysis_data(
        self, columns: List[str], symbols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        columns_ = ",".join(columns)
        if symbols:
            symbols_str = ",".join([f"'{s}'" for s in symbols])
            query = f"""SELECT {columns_} FROM analysis WHERE symbol IN ({symbols_str})"""  # noqa: S608
        else:
            query = f"""SELECT {columns_} FROM analysis"""  # noqa: S608
        return pd.read_sql_query(query, self._engine)

    def _read_filter_query(self, query: str) -> pd.DataFrame:
        return pd.read_sql(
            query,
            con=self._engine,
        )

    def _read_job_trackers(self) -> pd.DataFrame:
        query = "SELECT * FROM jobtracker ORDER BY started_at DESC"
        return pd.read_sql_query(query, self._engine)

    def write_job_tracker(self, job_tracker: JobTracker) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(JobTrackerORM)
                .prefix_with("OR REPLACE")
                .values(job_tracker.model_dump())
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def read_job_tracker(self, task_id: str) -> Optional[JobTracker]:
        stmt = select(JobTrackerORM).where(JobTrackerORM.job_id == task_id)
        with Session(self._engine) as session:
            result = session.execute(stmt).scalar_one_or_none()
            if result:
                return JobTracker.model_validate(result.model_dump())
            return None

    def delete_job_trackers(self, job_ids: List[str]) -> None:
        with Session(self._engine) as session:
            stmt = delete(JobTrackerORM).where(JobTrackerORM.job_id.in_(job_ids))  # type: ignore
            result = session.execute(stmt)

            if result.rowcount > 0:  # type: ignore
                session.commit()
            else:
                logger.warning(f"Job tracker(s) with ID(s) {job_ids} not found.")

    def update_job_tracker_status(self, job_tracker_status: JobTrackerStatus) -> None:
        with Session(self._engine) as session:
            stmt = (
                update(JobTrackerORM)
                .where(JobTrackerORM.job_id == job_tracker_status.job_id)  # type: ignore
                .values(status=job_tracker_status.status)
            )
            result = session.execute(stmt)

            if result.rowcount > 0:  # type: ignore
                session.commit()
            else:
                logger.warning(
                    f"Job tracker with ID {job_tracker_status.job_id} not found."
                )

    def read_filtered_results(self, name: str) -> Optional[FilteredResults]:
        with Session(self._engine) as session:
            stmt = select(FilteredResultsORM).where(FilteredResultsORM.name == name)
            result = session.execute(stmt).scalar_one_or_none()

            if result:
                return FilteredResults.model_validate(
                    result.model_dump()
                )  # if you're using Pydantic or DTOs
            return None

    def read_list_filtered_results(self) -> List[str]:
        with Session(self._engine) as session:
            stmt = select(FilteredResultsORM.name)
            result = session.execute(stmt).scalars().all()
            return list(result)

    def write_filtered_results(self, filtered_results: FilteredResults) -> None:
        with Session(self._engine) as session:
            data = filtered_results.model_dump_json(
                exclude_unset=True, exclude_defaults=True
            )
            stmt = (
                insert(FilteredResultsORM)
                .prefix_with("OR REPLACE")
                .values(json.loads(data))
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def read_dates(self, symbol: str) -> List[date]:
        with Session(self._engine) as session:
            return [
                r.date()
                for r in session.exec(
                    select(EarningsDateORM.date).where(EarningsDateORM.symbol == symbol)
                )
            ]

    def read_industry_symbols(
        self, industries: List[Industry], country: Country, source: Sources = "Yfinance"
    ) -> List[str]:
        with Session(self._engine) as session:
            stmt = select(EquityORM.symbol).where(
                EquityORM.industry.in_(industries),
                EquityORM.source == source,
                EquityORM.country == country,
            )
            result = session.exec(stmt).all()
            return list(result)

    def read_industry_group_symbols(
        self,
        industry_groups: List[IndustryGroup],
        country: Country,
        source: Sources = "Yfinance",
    ) -> List[str]:
        with Session(self._engine) as session:
            stmt = select(EquityORM.symbol).where(
                EquityORM.industry_group.in_(industry_groups),
                EquityORM.source == source,
                EquityORM.country == country,
            )
            result = session.exec(stmt).all()
            return list(result)

    def read_sector_symbols(
        self, sectors: List[Sector], country: Country, source: Sources = "Yfinance"
    ) -> List[str]:
        with Session(self._engine) as session:
            stmt = select(EquityORM.symbol).where(
                EquityORM.sector.in_(sectors),
                EquityORM.source == source,
                EquityORM.country == country,
            )
            result = session.exec(stmt).all()
            return list(result)

    def write_returns(self, industry_returns: List[IndustryView]) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(IndustryViewORM)
                .prefix_with("OR REPLACE")
                .values([a.model_dump() for a in industry_returns])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def read_returns(
        self, type: Type, industry: Industry, country: Country
    ) -> List[IndustryView]:
        with Session(self._engine) as session:
            stmt = select(IndustryViewORM)
            if industry:
                stmt = stmt.where(IndustryViewORM.industry == industry)
            if country:
                stmt = stmt.where(IndustryViewORM.country == country)
            result = session.exec(stmt).all()
            return [IndustryView.model_validate(r) for r in result]

    def write_signal_series(self, signal_series: List[SignalSeries]) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(SignalSeriesORM)
                .prefix_with("OR REPLACE")
                .values([a.model_dump() for a in signal_series])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def read_signal_series(
        self, name: str, start_date: date, end_date: date
    ) -> List[str]:
        with Session(self._engine) as session:
            stmt = select(SignalSeriesORM.symbol).where(
                SignalSeriesORM.name == name,
                SignalSeriesORM.date >= start_date,  # type: ignore
                SignalSeriesORM.date <= end_date,  # type: ignore
            )
            return list(set(session.exec(stmt).all()))

    def read_symbol_series(
        self, symbol: str, start_date: date, end_date: Optional[date] = None
    ) -> pd.DataFrame:

        with Session(self._engine) as session:
            query_ = select(PriceORM)
            query_ = query_.where(PriceORM.symbol == symbol)
            if end_date:
                query_ = query_.where(
                    PriceORM.date >= start_date, PriceORM.date <= end_date
                )
            else:
                query_ = query_.where(PriceORM.date >= start_date)
            series = session.exec(query_).all()
            prices = [Price.model_validate(serie) for serie in series]
            return Prices(prices=prices).to_dataframe()  # type: ignore

    def write_many_backtest_results(
        self, backtest_results: List["BacktestResult"]
    ) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(BacktestResultORM)
                .prefix_with("OR REPLACE")
                .values([a.model_dump() for a in backtest_results])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def read_many_backtest_results(
        self, query: Optional["BacktestResultQuery"] = None
    ) -> List["BacktestResult"]:
        from bullish.analysis.backtest import BacktestResult

        with Session(self._engine) as session:
            stmt = select(BacktestResultORM)
            results = session.exec(stmt).all()
            if results:
                return [BacktestResult.model_validate(r) for r in results]
            else:
                return []

    def read_next_earnings_date(self, symbol: str) -> Optional[date]:
        with Session(self._engine) as session:
            stmt = select(EarningsDateORM.date).where(
                EarningsDateORM.symbol == symbol, EarningsDateORM.date > date.today()
            )
            result = session.exec(stmt).first()
            if result:
                return result.date()  # type: ignore
            return None

    def read_subject(self, symbol: str) -> Optional[SubjectAnalysis]:
        sql = text(
            """
            SELECT *
            FROM   subject
            WHERE  symbol = :symbol
            ORDER  BY date DESC
            LIMIT  1
        """
        )
        sql_oai = text(
            """
            SELECT *
            FROM   openai
            WHERE  symbol = :symbol
            ORDER  BY news_date DESC
            LIMIT  1
        """
        )

        with Session(self._engine) as session:
            row = session.execute(sql, {"symbol": symbol}).mappings().one_or_none()
            row_oai = (
                session.execute(sql_oai, {"symbol": symbol}).mappings().one_or_none()
            )
            row_dict = {}
            if row:
                row_dict = dict(row)
                row_dict = row_dict | {"news_date": row_dict["date"]}
            if row_oai:
                row_dict_oai = dict(row_oai)
                row_dict = row_dict | {
                    "oai_news_date": row_dict_oai.get("news_date"),
                    "oai_recent_news": row_dict_oai.get("recent_news"),
                    "oai_recommendation": row_dict_oai.get("recommendation"),
                    "oai_explanation": row_dict_oai.get("explanation"),
                    "oai_high_price_target": row_dict_oai.get("high_price_target"),
                    "oai_low_price_target": row_dict_oai.get("low_price_target"),
                    "oai_moat": row_dict_oai.get("moat"),
                }

            return SubjectAnalysis.model_validate(row_dict)

    def write_many_openai_news(self, openai_news: List["OpenAINews"]) -> None:
        with Session(self._engine) as session:
            stmt = (
                insert(OpenAINewsORM)
                .prefix_with("OR REPLACE")
                .values([a.model_dump() for a in openai_news])
            )
            session.exec(stmt)  # type: ignore
            session.commit()

    def update_analysis(self, symbol: str, fields: Dict[str, Any]) -> None:
        with Session(self._engine) as session:
            stmt = (
                update(AnalysisORM).where(AnalysisORM.symbol == symbol).values(**fields)  # type: ignore
            )
            session.exec(stmt)  # type: ignore
            session.commit()



"""

Revision ID: fc191121f522
Revises: 49c83f9eb5ac
Create Date: 2025-07-06 17:33:02.231374

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "fc191121f522"
down_revision: Union[str, None] = "49c83f9eb5ac"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("lower_than_200_day_high", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("lower_than_20_day_high", sa.Date(), nullable=True)
        )
        batch_op.add_column(sa.Column("rsi_neutral", sa.Date(), nullable=True))
        batch_op.create_index(
            "ix_analysis_lower_than_200_day_high",
            ["lower_than_200_day_high"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_lower_than_20_day_high",
            ["lower_than_20_day_high"],
            unique=False,
        )
        batch_op.create_index("ix_analysis_rsi_neutral", ["rsi_neutral"], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_rsi_neutral")
        batch_op.drop_index("ix_analysis_lower_than_20_day_high")
        batch_op.drop_index("ix_analysis_lower_than_200_day_high")
        batch_op.drop_column("rsi_neutral")
        batch_op.drop_column("lower_than_20_day_high")
        batch_op.drop_column("lower_than_200_day_high")

    # ### end Alembic commands ###



import os
from pathlib import Path

from alembic import command
from alembic.config import Config
from alembic.util import CommandError

from bullish.database.scripts.stamp import stamp

DATABASE_PATH = Path(__file__).parents[3] / "tests" / "data" / "bear.db"


def upgrade(database_path: Path) -> None:
    root_folder = Path(__file__).parents[1]
    database_url = f"sqlite:///{database_path}"
    os.environ.update({"DATABASE_URL": database_url})
    alembic_cfg = Config(root_folder / "alembic" / "alembic.ini")
    alembic_cfg.set_main_option("script_location", str(root_folder / "alembic"))
    try:
        command.upgrade(alembic_cfg, "head")
    except CommandError:
        stamp(database_path)
        command.upgrade(alembic_cfg, "head")


if __name__ == "__main__":
    upgrade(DATABASE_PATH)



"""

Revision ID: cc28171c21a4
Revises: 260fcff7212e
Create Date: 2025-08-15 17:04:59.467407

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "cc28171c21a4"
down_revision: Union[str, None] = "260fcff7212e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(sa.Column("adx_14", sa.Date(), nullable=True))
        batch_op.add_column(sa.Column("adx_14_overbought", sa.Date(), nullable=True))
        batch_op.create_index("ix_analysis_adx_14", ["adx_14"], unique=False)
        batch_op.create_index(
            "ix_analysis_adx_14_overbought", ["adx_14_overbought"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_adx_14_overbought")
        batch_op.drop_index("ix_analysis_adx_14")
        batch_op.drop_column("adx_14_overbought")
        batch_op.drop_column("adx_14")

    # ### end Alembic commands ###



import abc
import logging
from datetime import date
from typing import List, Optional, Dict, Any

import pandas as pd
from bearish.interface.interface import BearishDbBase  # type: ignore
from bearish.models.base import Ticker  # type: ignore
from bearish.types import Sources  # type: ignore


from bullish.analysis.analysis import Analysis, AnalysisView, SubjectAnalysis
from bullish.analysis.backtest import BacktestResult, BacktestResultQuery
from bullish.analysis.constants import Industry, Sector, IndustryGroup, Country
from bullish.analysis.filter import FilterQuery, FilteredResults
from bullish.analysis.indicators import SignalSeries
from bullish.analysis.industry_views import Type, IndustryView
from bullish.analysis.openai import OpenAINews
from bullish.jobs.models import JobTracker, JobTrackerStatus, add_icons

logger = logging.getLogger(__name__)


class BullishDbBase(BearishDbBase):  # type: ignore
    def write_analysis(self, analysis: "Analysis") -> None:
        return self._write_analysis(analysis)

    def write_many_analysis(self, many_analysis: List["Analysis"]) -> None:
        return self._write_many_analysis(many_analysis)

    def read_analysis(self, ticker: Ticker) -> Optional["Analysis"]:
        return self._read_analysis(ticker)

    def read_filter_query(self, query: FilterQuery) -> pd.DataFrame:

        query_ = query.to_query()
        fields = ",".join(list(AnalysisView.model_fields))
        query_str: str = f""" 
        SELECT {fields} FROM analysis WHERE {query_}
        """  # noqa: S608
        return self._read_filter_query(query_str)

    def read_analysis_data(
        self, columns: Optional[List[str]] = None, symbols: Optional[List[str]] = None
    ) -> pd.DataFrame:
        columns = columns or list(AnalysisView.model_fields)
        data = self._read_analysis_data(columns, symbols=symbols)
        if set(data.columns) != set(columns):
            raise ValueError(
                f"Expected columns {columns}, but got {data.columns.tolist()}"
            )
        return data

    def read_job_trackers(self) -> pd.DataFrame:
        return add_icons(self._read_job_trackers())

    @abc.abstractmethod
    def _read_job_trackers(self) -> pd.DataFrame: ...

    @abc.abstractmethod
    def write_job_tracker(self, job_tracker: JobTracker) -> None: ...

    @abc.abstractmethod
    def delete_job_trackers(self, job_ids: List[str]) -> None: ...

    @abc.abstractmethod
    def update_job_tracker_status(
        self, job_tracker_status: JobTrackerStatus
    ) -> None: ...

    @abc.abstractmethod
    def _write_analysis(self, analysis: "Analysis") -> None: ...

    @abc.abstractmethod
    def _write_many_analysis(self, many_analysis: List["Analysis"]) -> None: ...

    @abc.abstractmethod
    def _read_analysis(self, ticker: Ticker) -> Optional["Analysis"]: ...

    @abc.abstractmethod
    def _read_filter_query(self, query: str) -> pd.DataFrame: ...

    @abc.abstractmethod
    def _read_analysis_data(
        self, columns: List[str], symbols: Optional[List[str]] = None
    ) -> pd.DataFrame: ...

    @abc.abstractmethod
    def read_filtered_results(self, name: str) -> Optional[FilteredResults]: ...

    @abc.abstractmethod
    def read_list_filtered_results(self) -> List[str]: ...

    @abc.abstractmethod
    def write_filtered_results(self, filtered_results: FilteredResults) -> None: ...

    @abc.abstractmethod
    def read_symbols(self) -> List[str]: ...

    @abc.abstractmethod
    def read_job_tracker(self, task_id: str) -> Optional[JobTracker]: ...

    @abc.abstractmethod
    def read_dates(self, symbol: str) -> List[date]: ...

    @abc.abstractmethod
    def read_industry_symbols(
        self, industries: List[Industry], country: Country, source: Sources = "Yfinance"
    ) -> List[str]: ...

    @abc.abstractmethod
    def read_industry_group_symbols(
        self,
        industry_groups: List[IndustryGroup],
        country: Country,
        source: Sources = "Yfinance",
    ) -> List[str]: ...

    @abc.abstractmethod
    def read_sector_symbols(
        self, sectors: List[Sector], country: Country, source: Sources = "Yfinance"
    ) -> List[str]: ...

    @abc.abstractmethod
    def write_returns(self, industry_returns: List[IndustryView]) -> None: ...

    @abc.abstractmethod
    def read_returns(
        self, type: Type, industry: Industry, country: Country
    ) -> List[IndustryView]: ...

    @abc.abstractmethod
    def write_signal_series(self, signal_series: List[SignalSeries]) -> None: ...

    @abc.abstractmethod
    def read_signal_series(
        self, name: str, start_date: date, end_date: date
    ) -> List[str]: ...

    @abc.abstractmethod
    def read_symbol_series(
        self, symbol: str, start_date: date, end_date: Optional[date] = None
    ) -> pd.DataFrame: ...
    @abc.abstractmethod
    def write_many_backtest_results(
        self, backtest_results: List[BacktestResult]
    ) -> None: ...

    @abc.abstractmethod
    def read_many_backtest_results(
        self, query: Optional[BacktestResultQuery] = None
    ) -> List[BacktestResult]: ...

    @abc.abstractmethod
    def read_next_earnings_date(self, symbol: str) -> Optional[date]: ...

    @abc.abstractmethod
    def read_subject(self, symbol: str) -> Optional[SubjectAnalysis]: ...
    @abc.abstractmethod
    def write_many_openai_news(self, openai_news: List[OpenAINews]) -> None: ...

    @abc.abstractmethod
    def update_analysis(self, symbol: str, fields: Dict[str, Any]) -> None: ...



import os
import sqlite3
from pathlib import Path

from alembic.config import Config

from bullish.database.settings import TEST_DATABASE_PATH


def stamp(database_path: Path) -> None:
    database_url = f"sqlite:///{database_path}"
    root_folder = Path(__file__).parents[1]
    os.environ.update({"DATABASE_URL": database_url})
    with sqlite3.connect(database_path) as conn:
        conn.execute("DROP TABLE IF EXISTS  alembic_version;")
        conn.execute("DROP TABLE IF EXISTS view;")
        conn.execute("DROP TABLE IF EXISTS analysis;")
        conn.execute("DROP TABLE IF EXISTS filteredresults;")
        conn.commit()
    alembic_cfg = Config(root_folder / "alembic" / "alembic.ini")
    alembic_cfg.set_main_option("script_location", str(root_folder / "alembic"))


if __name__ == "__main__":
    stamp(TEST_DATABASE_PATH)



"""

Revision ID: ae444f338124
Revises: d0e58e050845
Create Date: 2025-08-07 11:07:48.169941

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "ae444f338124"
down_revision: Union[str, None] = "d0e58e050845"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("consensus", sqlmodel.sql.sqltypes.AutoString(), nullable=True)
        )
        batch_op.add_column(
            sa.Column(
                "recommendation", sqlmodel.sql.sqltypes.AutoString(), nullable=True
            )
        )
        batch_op.add_column(sa.Column("high_price_target", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("low_price_target", sa.Float(), nullable=True))
        batch_op.add_column(
            sa.Column("reason", sqlmodel.sql.sqltypes.AutoString(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("explanation", sqlmodel.sql.sqltypes.AutoString(), nullable=True)
        )
        batch_op.add_column(sa.Column("news_date", sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column("news_summary", sa.JSON(), nullable=True))
        batch_op.alter_column(
            "summary",
            existing_type=sa.VARCHAR(),
            type_=sa.JSON(),
            existing_nullable=True,
        )
        batch_op.create_index("ix_analysis_consensus", ["consensus"], unique=False)
        batch_op.create_index("ix_analysis_explanation", ["explanation"], unique=False)
        batch_op.create_index(
            "ix_analysis_high_price_target", ["high_price_target"], unique=False
        )
        batch_op.create_index(
            "ix_analysis_low_price_target", ["low_price_target"], unique=False
        )
        batch_op.create_index("ix_analysis_news_date", ["news_date"], unique=False)
        batch_op.create_index(
            "ix_analysis_news_summary", ["news_summary"], unique=False
        )
        batch_op.create_index("ix_analysis_reason", ["reason"], unique=False)
        batch_op.create_index(
            "ix_analysis_recommendation", ["recommendation"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index("ix_analysis_recommendation")
        batch_op.drop_index("ix_analysis_reason")
        batch_op.drop_index("ix_analysis_news_summary")
        batch_op.drop_index("ix_analysis_news_date")
        batch_op.drop_index("ix_analysis_low_price_target")
        batch_op.drop_index("ix_analysis_high_price_target")
        batch_op.drop_index("ix_analysis_explanation")
        batch_op.drop_index("ix_analysis_consensus")
        batch_op.alter_column(
            "summary",
            existing_type=sa.JSON(),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )
        batch_op.drop_column("news_summary")
        batch_op.drop_column("news_date")
        batch_op.drop_column("explanation")
        batch_op.drop_column("reason")
        batch_op.drop_column("low_price_target")
        batch_op.drop_column("high_price_target")
        batch_op.drop_column("recommendation")
        batch_op.drop_column("consensus")

    # ### end Alembic commands ###



"""

Revision ID: 73564b60fe24
Revises: 4b0a2f40b7d3
Create Date: 2025-06-20 17:08:28.818293

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "73564b60fe24"
down_revision: Union[str, None] = "4b0a2f40b7d3"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "filteredresults",
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("filter_query", sa.JSON(), nullable=True),
        sa.Column("symbols", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("filteredresults")
    # ### end Alembic commands ###



import logging
from datetime import date
from typing import Optional, List, Callable, Any, Literal, Dict

import numpy as np
import pandas as pd
from pydantic import BaseModel, Field, PrivateAttr, create_model

from bullish.analysis.functions import (
    ADX,
    MACD,
    RSI,
    STOCH,
    MFI,
    ROC,
    CANDLESTOCK_PATTERNS,
    SMA,
    ADOSC,
    PRICE,
    cross_simple,
    cross_value_series,
    find_last_true_run_start,
    VOLUME,
)

logger = logging.getLogger(__name__)
SignalType = Literal["Short", "Long", "Oversold", "Overbought", "Value"]


def _last_date(d: pd.Series) -> Optional[date]:
    d_valid = d[d == 1]
    if d_valid.empty:
        return None
    last_index = d_valid.last_valid_index()
    return last_index.date() if last_index is not None else None  # type: ignore


class ProcessingFunction(BaseModel):
    date: Callable[[pd.Series], Optional[date]] = Field(default=_last_date)
    number: Callable[[pd.Series], Optional[float]] = Field(
        default=lambda d: d.iloc[-1] if not d.dropna().empty else None
    )


class SignalSeries(BaseModel):
    name: str
    date: date
    value: float
    symbol: str


class Signal(BaseModel):
    name: str
    type_info: SignalType
    type: Any
    range: Optional[List[float]] = None
    function: Callable[[pd.DataFrame], pd.Series]
    processing: ProcessingFunction = Field(default_factory=ProcessingFunction)
    description: str
    date: Optional[date] = None
    value: Optional[float] = None
    in_use_backtest: bool = False

    def is_date(self) -> bool:
        if self.type == Optional[date]:
            return True
        elif self.type == Optional[float]:
            return False
        else:
            raise NotImplementedError

    def apply_function(self, data: pd.DataFrame) -> pd.Series:
        result = self.function(data)
        if not isinstance(result, pd.Series):
            raise ValueError(
                f"Function for signal {self.name} must return a pandas Series"
            )
        return result

    def compute(self, data: pd.DataFrame) -> None:
        if self.is_date():
            self.date = self.processing.date(self.apply_function(data))
        else:
            self.value = self.processing.number(self.apply_function(data))

    def compute_series(self, data: pd.DataFrame) -> pd.Series:
        return self.apply_function(data)


class Indicator(BaseModel):
    name: str
    description: str
    expected_columns: List[str]
    function: Callable[[pd.DataFrame], pd.DataFrame]
    _data: pd.DataFrame = PrivateAttr(default=pd.DataFrame())
    signals: List[Signal] = Field(default_factory=list)

    def compute(self, data: pd.DataFrame) -> None:
        results = self.function(data)
        if not set(self.expected_columns).issubset(results.columns):
            raise ValueError(
                f"Expected columns {self.expected_columns}, but got {results.columns.tolist()}"
            )
        self._data = results
        self.compute_signals()

    def compute_signals(self) -> None:
        for signal in self.signals:
            try:
                signal.compute(self._data)
            except Exception as e:  # noqa: PERF203
                logger.error(
                    f"Fail to compute signal {signal.name} for indicator {self.name}: {e}"
                )

    def compute_series(self, data: pd.DataFrame, symbol: str) -> pd.DataFrame:
        series = []
        try:
            results = self.function(data)
        except Exception as e:
            logger.error(
                f"Failed to compute indicator {self.name} for symbol {symbol}: {e}"
            )
            return pd.DataFrame()
        if not set(self.expected_columns).issubset(results.columns):
            raise ValueError(
                f"Expected columns {self.expected_columns}, but got {results.columns.tolist()}"
            )
        for signal in self.signals:
            if not signal.in_use_backtest:
                continue
            try:
                series_ = signal.compute_series(results)
                if signal.type == Optional[date]:
                    series__ = pd.DataFrame(series_[series_ == 1].rename("value"))
                else:
                    series__ = pd.DataFrame(
                        series_[series_ != None].rename("value")  # noqa: E711
                    )

                series__["name"] = signal.name
                series__["date"] = series__.index.date  # type: ignore
                series__["symbol"] = symbol
                series__ = series__.reset_index(drop=True)
                series.append(series__)
            except Exception as e:
                logger.error(
                    f"Fail to compute signal {signal.name} for indicator {self.name}: {e}"
                )
        if not series:
            return pd.DataFrame()
        data = pd.concat(series).reset_index(drop=True)
        return data


def indicators_factory() -> List[Indicator]:
    return [
        Indicator(
            name="ADX_14",
            description="Average Directional Movement Index",
            expected_columns=["ADX_14", "MINUS_DI", "PLUS_DI"],
            function=ADX.call,
            signals=[
                Signal(
                    name="ADX_14_LONG",
                    description="ADX 14 Long Signal",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: (d.ADX_14 > 20) & (d.PLUS_DI > d.MINUS_DI),
                ),
                Signal(
                    name="ADX_14_SHORT",
                    description="ADX 14 Short Signal",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: (d.ADX_14 > 20) & (d.MINUS_DI > d.PLUS_DI),
                ),
                Signal(
                    name="ADX_14",
                    description="ADX 14",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: (d.ADX_14 > 25),
                ),
                Signal(
                    name="ADX_14_OVERBOUGHT",
                    description="ADX 14 OVERBOUGHT",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: (d.ADX_14 > 50),
                ),
            ],
        ),
        Indicator(
            name="MACD_12_26_9",
            description="Moving Average Convergence/Divergence",
            expected_columns=[
                "MACD_12_26_9",
                "MACD_12_26_9_SIGNAL",
                "MACD_12_26_9_HIST",
            ],
            function=MACD.call,
            signals=[
                Signal(
                    name="MACD_12_26_9_BULLISH_CROSSOVER",
                    description="MACD 12-26-9 Bullish Crossover",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_simple(
                        d.MACD_12_26_9, d.MACD_12_26_9_SIGNAL
                    ),
                    in_use_backtest=True,
                ),
                Signal(
                    name="MACD_12_26_9_BEARISH_CROSSOVER",
                    description="MACD 12-26-9 Bearish Crossover",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: cross_simple(
                        d.MACD_12_26_9_SIGNAL, d.MACD_12_26_9
                    ),
                ),
                Signal(
                    name="MACD_12_26_9_ZERO_LINE_CROSS_UP",
                    description="MACD 12-26-9 Zero Line Cross Up",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value_series(d.MACD_12_26_9, 0),
                ),
                Signal(
                    name="MACD_12_26_9_ZERO_LINE_CROSS_DOWN",
                    description="MACD 12-26-9 Zero Line Cross Down",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value_series(
                        d.MACD_12_26_9, 0, above=False
                    ),
                ),
            ],
        ),
        Indicator(
            name="RSI",
            description="Relative Strength Index",
            expected_columns=RSI.expected_columns,
            function=RSI.call,
            signals=[
                Signal(
                    name="RSI_BULLISH_CROSSOVER_30",
                    description="RSI Bullish Crossover",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value_series(d.RSI, 30),
                    in_use_backtest=True,
                ),
                Signal(
                    name="RSI_BULLISH_CROSSOVER_40",
                    description="RSI Bullish Crossover 40",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value_series(d.RSI, 40),
                    in_use_backtest=True,
                ),
                Signal(
                    name="RSI_BULLISH_CROSSOVER_45",
                    description="RSI Bullish Crossover 45",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: cross_value_series(d.RSI, 45),
                    in_use_backtest=True,
                ),
                Signal(
                    name="RSI_BEARISH_CROSSOVER",
                    description="RSI Bearish Crossover",
                    type_info="Short",
                    type=Optional[date],
                    function=lambda d: cross_value_series(d.RSI, 70, above=False),
                ),
                Signal(
                    name="RSI_OVERSOLD",
                    description="RSI Oversold Signal",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: (d.RSI <= 30) & (d.RSI > 0),
                    in_use_backtest=True,
                ),
                Signal(
                    name="RSI_OVERBOUGHT",
                    description="RSI Overbought Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: (d.RSI < 100) & (d.RSI > 70),
                ),
                Signal(
                    name="RSI_NEUTRAL",
                    description="RSI Neutral Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: (d.RSI < 60) & (d.RSI > 30),
                ),
                Signal(
                    name="RSI",
                    description="RSI value",
                    type_info="Overbought",
                    type=Optional[float],
                    function=lambda d: d.RSI,
                ),
            ],
        ),
        Indicator(
            name="STOCH",
            description="Stochastic",
            expected_columns=["SLOW_K", "SLOW_D"],
            function=STOCH.call,
            signals=[
                Signal(
                    name="STOCH_OVERSOLD",
                    description="Stoch Oversold Signal",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: (d.SLOW_K < 20) & (d.SLOW_K > 0),
                ),
                Signal(
                    name="STOCH_OVERBOUGHT",
                    description="Stoch Overbought Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: (d.SLOW_K < 100) & (d.SLOW_K > 80),
                ),
            ],
        ),
        Indicator(
            name="MFI",
            description="Money Flow Index",
            expected_columns=["MFI"],
            function=MFI.call,
            signals=[
                Signal(
                    name="MFI_OVERSOLD",
                    description="MFI Oversold Signal",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: (d.MFI < 20),
                ),
                Signal(
                    name="MFI_OVERBOUGHT",
                    description="MFI Overbought Signal",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: (d.MFI > 80),
                ),
            ],
        ),
        Indicator(
            name="SMA",
            description="Money Flow Index",
            expected_columns=["SMA_50", "SMA_200"],
            function=SMA.call,
            signals=[
                Signal(
                    name="GOLDEN_CROSS",
                    description="Golden cross: SMA 50 crosses above SMA 200",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: cross_simple(d.SMA_50, d.SMA_200),
                    in_use_backtest=True,
                ),
                Signal(
                    name="DEATH_CROSS",
                    description="Death cross: SMA 50 crosses below SMA 200",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: cross_simple(d.SMA_50, d.SMA_200, above=False),
                ),
                Signal(
                    name="SMA_50_ABOVE_SMA_200",
                    description="SMA 50 is above SMA 200",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d.SMA_50 > d.SMA_200,
                    in_use_backtest=True,
                    processing=ProcessingFunction(date=find_last_true_run_start),
                ),
                Signal(
                    name="SMA_50_BELOW_SMA_200",
                    description="SMA 50 is below SMA 200",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d.SMA_50 < d.SMA_200,
                    in_use_backtest=True,
                    processing=ProcessingFunction(date=find_last_true_run_start),
                ),
                Signal(
                    name="PRICE_ABOVE_SMA_50",
                    description="Price is above SMA 50",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d.SMA_50 < d.CLOSE,
                    in_use_backtest=True,
                    processing=ProcessingFunction(date=find_last_true_run_start),
                ),
                Signal(
                    name="PRICE_BELOW_SMA_50",
                    description="Price is below SMA 50",
                    type_info="Overbought",
                    type=Optional[date],
                    function=lambda d: d.SMA_50 > d.CLOSE,
                    in_use_backtest=True,
                    processing=ProcessingFunction(date=find_last_true_run_start),
                ),
            ],
        ),
        Indicator(
            name="PRICE",
            description="Price based indicators",
            expected_columns=PRICE.expected_columns,
            function=PRICE.call,
            signals=[
                Signal(
                    name="LOWER_THAN_200_DAY_HIGH",
                    description="Current price is lower than the 200-day high",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: 0.6 * d["200_DAY_HIGH"] > d.LAST_PRICE,
                ),
                Signal(
                    name="LOWER_THAN_20_DAY_HIGH",
                    description="Current price is lower than the 20-day high",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: 0.6 * d["20_DAY_HIGH"] > d.LAST_PRICE,
                ),
                Signal(
                    name="WEEKLY_GROWTH",
                    description="weekly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: d.WEEKLY_GROWTH,
                ),
                Signal(
                    name="MONTHLY_GROWTH",
                    description="Median monthly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: d.MONTHLY_GROWTH,
                ),
                Signal(
                    name="YEARLY_GROWTH",
                    description="Median yearly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: d.YEARLY_GROWTH,
                ),
                Signal(
                    name="MEDIAN_WEEKLY_GROWTH",
                    description="Median weekly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: d.WEEKLY_GROWTH,
                    processing=ProcessingFunction(
                        number=lambda v: np.median(v.unique())
                    ),
                ),
                Signal(
                    name="MEDIAN_MONTHLY_GROWTH",
                    description="Median monthly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: d.MONTHLY_GROWTH,
                    processing=ProcessingFunction(
                        number=lambda v: np.median(v.unique())
                    ),
                ),
                Signal(
                    name="MEDIAN_YEARLY_GROWTH",
                    description="Median yearly growth",
                    type_info="Oversold",
                    type=Optional[float],
                    function=lambda d: d.YEARLY_GROWTH,
                    processing=ProcessingFunction(
                        number=lambda v: np.median(v.unique())
                    ),
                ),
                Signal(
                    name="LOWER_THAN_20_DAY_HIGH",
                    description="Current price is lower than the 20-day high",
                    type_info="Oversold",
                    type=Optional[date],
                    function=lambda d: 0.6 * d["20_DAY_HIGH"] > d.LAST_PRICE,
                ),
            ],
        ),
        Indicator(
            name="VOLUME",
            description="Volume based indicators",
            expected_columns=VOLUME.expected_columns,
            function=VOLUME.call,
            signals=[
                Signal(
                    name="AVERAGE_VOLUME_10",
                    type_info="Value",
                    description="Average volume over the last 10 days",
                    type=Optional[float],
                    function=lambda d: d.AVERAGE_VOLUME_10,
                ),
                Signal(
                    name="AVERAGE_VOLUME_30",
                    type_info="Value",
                    description="Average volume over the last 30 days",
                    type=Optional[float],
                    function=lambda d: d.AVERAGE_VOLUME_30,
                ),
                Signal(
                    name="VOLUME_ABOVE_AVERAGE",
                    type_info="Value",
                    description="Volume above average volume over the last 30 days",
                    type=Optional[date],
                    function=lambda d: d.AVERAGE_VOLUME_30 < d.VOLUME,
                ),
            ],
        ),
        Indicator(
            name="ROC",
            description="Rate Of Change",
            expected_columns=ROC.expected_columns,
            function=ROC.call,
            signals=[
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_1",
                    type_info="Value",
                    description="Median daily Rate of Change of the last 30 days",
                    type=Optional[float],
                    function=lambda d: d.ROC_1,
                    processing=ProcessingFunction(
                        number=lambda v: np.median(v.tolist()[-30:])
                    ),
                ),
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_7_4",
                    type_info="Value",
                    description="Median weekly Rate of Change of the last 4 weeks",
                    type=Optional[float],
                    function=lambda d: d.ROC_7,
                    processing=ProcessingFunction(
                        number=lambda v: np.median(v.tolist()[-4:])
                    ),
                ),
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_7_12",
                    type_info="Value",
                    description="Median weekly Rate of Change of the last 12 weeks",
                    type=Optional[float],
                    function=lambda d: d.ROC_7,
                    processing=ProcessingFunction(
                        number=lambda v: np.median(v.tolist()[-12:])
                    ),
                ),
                Signal(
                    name="MEDIAN_RATE_OF_CHANGE_30",
                    type_info="Value",
                    description="Median monthly Rate of Change of the last 12 Months",
                    type=Optional[float],
                    function=lambda d: d.ROC_30,
                    processing=ProcessingFunction(
                        number=lambda v: np.median(v.tolist()[-12:])
                    ),
                ),
                Signal(
                    name="RATE_OF_CHANGE_30",
                    type_info="Value",
                    description="30-day Rate of Change",
                    type=Optional[float],
                    function=lambda d: d.ROC_30,
                ),
                Signal(
                    name="RATE_OF_CHANGE_7",
                    type_info="Value",
                    description="7-day Rate of Change",
                    type=Optional[float],
                    function=lambda d: d.ROC_7,
                ),
                Signal(
                    name="MOMENTUM",
                    type_info="Value",
                    description="7-day Rate of Change",
                    type=Optional[float],
                    function=lambda d: d.MOM,
                ),
            ],
        ),
        Indicator(
            name="ADOSC",
            description="Chaikin A/D Oscillator",
            expected_columns=["ADOSC", "ADOSC_SIGNAL"],
            function=ADOSC.call,
            signals=[
                Signal(
                    name="ADOSC_CROSSES_ABOVE_0",
                    type_info="Oversold",
                    description="Bullish momentum in money flow",
                    type=Optional[date],
                    function=lambda d: cross_value_series(d.ADOSC, 0, above=True),
                ),
                Signal(
                    name="POSITIVE_ADOSC_20_DAY_BREAKOUT",
                    type_info="Oversold",
                    description="20-day breakout confirmed by positive ADOSC",
                    type=Optional[date],
                    function=lambda d: (d.ADOSC_SIGNAL == True),  # noqa: E712
                ),
            ],
        ),
        Indicator(
            name="CANDLESTICKS",
            description="Candlestick Patterns",
            expected_columns=[
                "CDLMORNINGSTAR",
                "CDL3LINESTRIKE",
                "CDL3WHITESOLDIERS",
                "CDLABANDONEDBABY",
                "CDLTASUKIGAP",
                "CDLPIERCING",
                "CDLENGULFING",
            ],
            function=CANDLESTOCK_PATTERNS.call,
            signals=[
                Signal(
                    name="CDLMORNINGSTAR",
                    type_info="Long",
                    description="Morning Star Candlestick Pattern",
                    type=Optional[date],
                    function=lambda d: d.CDLMORNINGSTAR == 100,
                ),
                Signal(
                    name="CDL3LINESTRIKE",
                    description="3 Line Strike Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d.CDL3LINESTRIKE == 100,
                ),
                Signal(
                    name="CDL3WHITESOLDIERS",
                    description="3 White Soldiers Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d.CDL3WHITESOLDIERS == 100,
                ),
                Signal(
                    name="CDLABANDONEDBABY",
                    description="Abandoned Baby Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d.CDLABANDONEDBABY == 100,
                ),
                Signal(
                    name="CDLTASUKIGAP",
                    description="Tasukigap Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d.CDLTASUKIGAP == 100,
                ),
                Signal(
                    name="CDLPIERCING",
                    description="Piercing Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d.CDLPIERCING == 100,
                ),
                Signal(
                    name="CDLENGULFING",
                    description="Engulfing Candlestick Pattern",
                    type_info="Long",
                    type=Optional[date],
                    function=lambda d: d.CDLENGULFING == 100,
                ),
            ],
        ),
    ]


class Indicators(BaseModel):
    indicators: List[Indicator] = Field(default_factory=indicators_factory)

    def in_use_backtest(self) -> List[str]:
        return [
            signal.name.lower()
            for indicator in self.indicators
            for signal in indicator.signals
            if signal.in_use_backtest
        ]

    def _compute(self, data: pd.DataFrame) -> None:
        for indicator in self.indicators:
            try:
                indicator.compute(data)
            except Exception as e:
                logger.error(f"Failed to compute indicator {indicator.name}: {e}")
                continue
            logger.info(
                f"Computed {indicator.name} with {len(indicator.signals)} signals"
            )

    def compute_series(self, data: pd.DataFrame, symbol: str) -> List[SignalSeries]:
        data__ = pd.concat(
            [indicator.compute_series(data, symbol) for indicator in self.indicators]
        )
        return [
            SignalSeries.model_validate(s) for s in data__.to_dict(orient="records")
        ]

    def compute(self, data: pd.DataFrame) -> Dict[str, Any]:
        self._compute(data)
        res = {}
        for indicator in self.indicators:
            for signal in indicator.signals:
                res[signal.name.lower()] = (
                    signal.date if signal.is_date() else signal.value
                )
        return res

    def create_indicator_models(self) -> List[type[BaseModel]]:
        models = []
        for indicator in self.indicators:
            model_parameters = {}
            for signal in indicator.signals:
                range_ = {}
                if signal.range:
                    range_ = {"ge": signal.range[0], "le": signal.range[1]}
                model_parameters[signal.name.lower()] = (
                    signal.type,
                    Field(  # type: ignore
                        None,
                        **range_,
                        description=(
                            signal.description
                            or " ".join(signal.name.lower().capitalize().split("_"))
                        ),
                    ),
                )
            model = create_model(indicator.name, **model_parameters)  # type: ignore
            model._description = indicator.description
            models.append(model)
        return models


IndicatorModels = Indicators().create_indicator_models()



import json
import logging
import random
from datetime import date, timedelta
from io import StringIO
from typing import TYPE_CHECKING, Optional, Union, List, Dict, Any

import numpy as np
import pandas as pd
from pydantic import BaseModel, Field, model_validator


import plotly.graph_objects as go

if TYPE_CHECKING:
    from bullish.analysis.predefined_filters import NamedFilterQuery
    from bullish.database.crud import BullishDb

logger = logging.getLogger(__name__)
COLOR = {
    "mean": "#1f77b4",  # A refined blue (Plotly default)
    "upper": "#d62728",  # Strong red
    "lower": "#2ca02c",  # Rich green
    "median": "#ff7f0e",  # Bright orange
}


class BacktestQueryBase(BaseModel):
    name: str
    table: str


class BacktestQueryDate(BacktestQueryBase):

    start: date
    end: date


class BacktestQueryRange(BacktestQueryBase):

    min: float
    max: float


class BacktestQuerySelection(BacktestQueryBase):

    selections: List[str]

    def to_selections(self) -> str:
        return ", ".join([f"'{s}'" for s in self.selections])


class BacktestQueries(BaseModel):
    queries: list[Union[BacktestQueryDate, BacktestQueryRange, BacktestQuerySelection]]

    def to_query(self) -> str:
        query_parts = []
        for query in self.queries:
            if isinstance(query, (BacktestQueryDate)):
                query_parts.append(
                    f"SELECT symbol FROM {query.table} WHERE name='{query.name}' "  # noqa: S608
                    f"AND date >='{query.start}' AND date <='{query.end}'"
                )
            if isinstance(query, (BacktestQueryRange)):
                query_parts.append(
                    f"SELECT symbol FROM {query.table} WHERE "  # noqa: S608
                    f"{query.name} >= {query.min} AND {query.name} <= {query.max}"
                )
            if isinstance(query, (BacktestQuerySelection)):
                query_parts.append(
                    f"SELECT symbol FROM {query.table} WHERE "  # noqa: S608
                    f"{query.name} IN ({query.to_selections()})"
                )

        if len(query_parts) == 1:
            return query_parts[0]
        else:
            return " INTERSECT ".join(query_parts)


class ReturnPercentage(BaseModel):
    return_percentage: float = Field(
        default=12, description="Return percentage of the backtest"
    )


class BaseBacktestResult(BaseModel):
    start: date = Field(default=date.today() - timedelta(days=252))
    end: date = Field(default=date.today())
    investment: float = Field(default=1000)
    holding_period: int = Field(default=30 * 3)
    extend_days: int = Field(
        default=5,
        description="Extend the backtest by this many days if no symbols are found",
    )
    percentage: int = Field(default=12, description="Return percentage of the backtest")
    iterations: int = Field(default=200, description="Number of iterations to run")


class BacktestResultQuery(BaseBacktestResult):
    name: str


class BacktestResult(BacktestResultQuery):
    data: Dict[str, Any]

    def to_dataframe(self) -> pd.DataFrame:
        return pd.read_json(StringIO(json.dumps(self.data))).sort_index()


class BacktestResults(BaseModel):
    results: List[BacktestResult]

    def figure(self, type: str = "mean") -> go.Figure:
        fig = go.Figure()
        for result in self.results:
            data = result.to_dataframe()[type]
            fig.add_trace(
                go.Scatter(
                    x=data.index,
                    y=data,
                    mode="lines",
                    name=f"{result.name} ({type})",
                    line={"width": 1},
                    hovertemplate=(
                        "Date: %{x}<br>"
                        + "Price: %{y:.2f}<br>"
                        + f"Percentage: {result.percentage}<br>"
                        + f"Iterations: {result.iterations}<br>"
                        + f"Investment: {result.investment}<extra></extra>"
                    ),
                )
            )
        fig.update_layout(
            height=800,
            showlegend=True,
            margin={"t": 60, "b": 40},
        )

        return fig


class BackTestConfig(BaseBacktestResult):
    exit_strategy: ReturnPercentage = Field(default=ReturnPercentage)

    def to_base_backtest_result(self) -> BaseBacktestResult:
        return BaseBacktestResult(
            start=self.start,
            end=self.end,
            investment=self.investment,
            holding_period=self.holding_period,
            extend_days=self.extend_days,
            percentage=self.percentage,
            iterations=self.iterations,
        )


class Equity(BaseModel):
    symbol: str
    start: date
    end: date
    buy: float
    sell: float
    investment_in: float
    investment_out: Optional[float] = None

    def profit(self) -> float:
        return (self.sell - self.buy) * (self.investment_in / self.buy)

    def current_value(self) -> float:
        return self.investment_in + self.profit()

    def set_investment_out(self) -> None:
        self.investment_out = self.current_value()


class BackTest(BaseModel):
    equities: list[Equity] = Field(
        default_factory=list, description="List of equities bought during the backtest"
    )
    end: date = Field(default=date.today(), description="End date of the backtest")

    def valid(self) -> bool:
        return bool(self.equities)

    def total_profit(self) -> float:
        return sum(equity.profit() for equity in self.equities)

    def symbols(self) -> list[str]:
        return [equity.symbol for equity in self.equities]

    def show(self) -> None:
        for eq in self.equities:
            print(
                f"\n{eq.symbol} ({eq.type}): {eq.start}:{eq.investment_in} ({eq.buy}) - "
                f"{eq.end}:{eq.investment_out} ({eq.sell})"
            )

    def to_dataframe(self) -> pd.DataFrame:
        prices = [
            self.equities[0].investment_in,
            *[e.investment_out for e in self.equities],
        ]
        symbols = [self.equities[0].symbol, *[e.symbol for e in self.equities]]
        index = [self.equities[0].start, *[e.end for e in self.equities]]
        buy = [self.equities[0].buy, *[e.buy for e in self.equities]]
        sell = [self.equities[0].sell, *[e.sell for e in self.equities]]
        data = pd.DataFrame(
            np.array([prices, symbols, buy, sell]).T,
            index=index,
            columns=["prices", "symbols", "buy", "sell"],
        )
        data = data[~data.index.duplicated(keep="first")]
        return data

    def __hash__(self) -> int:
        return hash(tuple(sorted(equity.symbol for equity in self.equities)))


class BackTests(BaseModel):
    tests: list[BackTest] = Field(default_factory=list, description="List of backtests")
    config: BackTestConfig
    name: str

    @model_validator(mode="after")
    def _validate(self) -> "BackTests":
        self.tests = list(set(self.tests))  # Remove duplicates
        return self

    def to_dataframe(self) -> pd.DataFrame:

        data = (
            pd.concat([t.to_dataframe() for t in self.tests if t.valid()], axis=1)
            .sort_index()
            .fillna(method="ffill")
        )
        data = data[~data.index.duplicated(keep="first")]
        return data

    def to_error(self) -> pd.DataFrame:
        data_ = self.to_dataframe()
        mean = data_.prices.astype(float).mean(axis=1).rename("mean")
        std = data_.prices.astype(float).std(axis=1)
        median = data_.prices.astype(float).median(axis=1).rename("median")
        upper = (mean + std).rename("upper")
        lower = (mean - std).rename("lower")
        return pd.concat([mean, upper, lower, median], axis=1).sort_index()

    def to_backtest_result(self) -> BacktestResult:

        return BacktestResult.model_validate(
            self.config.to_base_backtest_result().model_dump()
            | {"data": json.loads(self.to_error().to_json()), "name": self.name}
        )

    def to_figure(self) -> go.Figure:

        data_ = self.to_dataframe()
        self.to_error()
        column_chunks = [data_.iloc[:, i : i + 4] for i in range(0, data_.shape[1], 4)]
        fig = go.Figure()
        for data in column_chunks:
            fig.add_trace(
                go.Scatter(
                    x=data.index,
                    y=data.prices.astype(float),
                    mode="lines",
                    showlegend=False,
                    customdata=data[
                        ["symbols", "sell", "buy"]
                    ],  # Include multiple overlay columns
                    line={"color": "grey", "width": 0.5},  # normal grey
                    opacity=0.5,
                    hovertemplate=(
                        "Date: %{x}<br>"
                        + "Price: %{y:.2f}<br>"
                        + "Symbols: %{customdata[0]}<br>"
                        + "Sell: %{customdata[1]}<br>"
                        + "Buy: %{customdata[2]}<extra></extra>"
                    ),
                )
            )
        for name, column in self.to_error().items():
            fig.add_trace(
                go.Scatter(
                    x=column.index,
                    y=column,
                    mode="lines",
                    line={"color": COLOR[name], "width": 1},
                    showlegend=True,
                    name=name,
                )
            )
        fig.update_layout(
            title="Predefined filter performance",
            xaxis_title="Date",
            yaxis_title="Prices [Currency]",
        )
        fig.show()
        return fig


def run_backtest(  # noqa: C901, PLR0915
    bullish_db: "BullishDb", named_filter: "NamedFilterQuery", config: BackTestConfig
) -> BackTest:
    equities = []
    start_date = config.start
    presence_delta = timedelta(days=config.holding_period)
    investment = config.investment
    exclude_symbols = []
    while True:
        symbols = []
        while not symbols:
            symbols = named_filter.get_backtesting_symbols(bullish_db, start_date)
            symbols = [b for b in symbols if b not in exclude_symbols]
            if symbols:
                break
            start_date = start_date + timedelta(days=config.extend_days)
            if start_date > config.end:
                logger.debug("No symbols found for the given date range.")
                break
        if symbols:
            symbol = random.choice(symbols)  # noqa: S311
            logger.debug(f"Found symbol: {symbol}, for date: {start_date}")
            enter_position = start_date
            end_position = None
            counter = 0
            buy_price = None
            while True:

                data = bullish_db.read_symbol_series(
                    symbol,
                    start_date=enter_position + counter * presence_delta,
                    end_date=enter_position + (counter + 1) * presence_delta,
                )
                if data.empty:
                    logger.debug(f"No data found for symbol: {symbol}")
                    exclude_symbols.append(symbol)
                    end_position = start_date
                    break
                data.index = data.index.tz_localize(None)
                if counter == 0:
                    enter_position_timestamp = data.close.first_valid_index()
                    enter_position = enter_position_timestamp.date()
                    buy_price = data.close.loc[enter_position_timestamp]

                mask = data.close >= buy_price * (
                    1 + config.percentage / (100 * (counter + 1))
                )
                mask_ = mask[mask == True]  # noqa: E712

                if mask_.empty:
                    if enter_position + (counter + 1) * presence_delta > config.end:
                        end_position = data.close.index[-1].date()
                        sell_price = data.close.iloc[-1]
                        equity = Equity(
                            symbol=symbol,
                            start=enter_position,
                            end=end_position,
                            buy=buy_price,
                            sell=sell_price,
                            investment_in=investment,
                        )
                        equity.set_investment_out()
                        equities.append(equity)
                        investment = equity.current_value()
                        end_position = config.end
                        break
                    counter += 1
                    continue
                else:
                    end_position_timestamp = data[mask].first_valid_index()
                    end_position = end_position_timestamp.date()
                    equity = Equity(
                        symbol=symbol,
                        start=enter_position,
                        end=end_position,
                        buy=buy_price,
                        sell=data[mask].close.loc[end_position_timestamp],
                        investment_in=investment,
                    )
                    equity.set_investment_out()
                    equities.append(equity)
                    investment = equity.current_value()
                    break

            start_date = end_position
        if start_date >= config.end:
            break
    back_test = BackTest(equities=equities)
    return back_test


def run_tests(
    bullish_db: "BullishDb", named_filter: "NamedFilterQuery", config: BackTestConfig
) -> BackTests:
    return BackTests(
        config=config,
        name=named_filter.name,
        tests=[
            run_backtest(bullish_db, named_filter, config)
            for _ in range(config.iterations)
        ],
    )


def run_many_tests(
    bullish_db: "BullishDb",
    named_filters: List["NamedFilterQuery"],
    config: BackTestConfig,
) -> None:
    back_tests = []
    for named_filter in named_filters:
        try:
            back_tests.append(
                run_tests(bullish_db, named_filter, config).to_backtest_result()
            )
        except Exception as e:  # noqa: PERF203
            logger.error(e)
            continue

    if back_tests:
        bullish_db.write_many_backtest_results(back_tests)



"""

Revision ID: 037dbd721317
Revises:
Create Date: 2025-06-14 16:50:56.919222

"""

from typing import Sequence, Union

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "037dbd721317"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "analysis",
        sa.Column("quarterly_positive_free_cash_flow", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_operating_cash_flow", sa.Float(), nullable=True),
        sa.Column(
            "quarterly_operating_cash_flow_is_higher_than_net_income",
            sa.Float(),
            nullable=True,
        ),
        sa.Column("quarterly_mean_capex_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_max_capex_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_min_capex_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_mean_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_max_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_min_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_net_income", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_operating_income", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_net_income", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_operating_income", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_diluted_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_basic_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_basic_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_diluted_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_debt_to_equity", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_return_on_assets", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_return_on_equity", sa.Float(), nullable=True),
        sa.Column("quarterly_earning_per_share", sa.Float(), nullable=True),
        sa.Column("positive_free_cash_flow", sa.Float(), nullable=True),
        sa.Column("growing_operating_cash_flow", sa.Float(), nullable=True),
        sa.Column(
            "operating_cash_flow_is_higher_than_net_income", sa.Float(), nullable=True
        ),
        sa.Column("mean_capex_ratio", sa.Float(), nullable=True),
        sa.Column("max_capex_ratio", sa.Float(), nullable=True),
        sa.Column("min_capex_ratio", sa.Float(), nullable=True),
        sa.Column("mean_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("max_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("min_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("positive_net_income", sa.Float(), nullable=True),
        sa.Column("positive_operating_income", sa.Float(), nullable=True),
        sa.Column("growing_net_income", sa.Float(), nullable=True),
        sa.Column("growing_operating_income", sa.Float(), nullable=True),
        sa.Column("positive_diluted_eps", sa.Float(), nullable=True),
        sa.Column("positive_basic_eps", sa.Float(), nullable=True),
        sa.Column("growing_basic_eps", sa.Float(), nullable=True),
        sa.Column("growing_diluted_eps", sa.Float(), nullable=True),
        sa.Column("positive_debt_to_equity", sa.Float(), nullable=True),
        sa.Column("positive_return_on_assets", sa.Float(), nullable=True),
        sa.Column("positive_return_on_equity", sa.Float(), nullable=True),
        sa.Column("earning_per_share", sa.Float(), nullable=True),
        sa.Column("rsi_last_value", sa.Float(), nullable=True),
        sa.Column("macd_12_26_9_buy_date", sa.Date(), nullable=True),
        sa.Column("ma_50_200_buy_date", sa.Date(), nullable=True),
        sa.Column("slope_7", sa.Float(), nullable=True),
        sa.Column("slope_14", sa.Float(), nullable=True),
        sa.Column("slope_30", sa.Float(), nullable=True),
        sa.Column("slope_60", sa.Float(), nullable=True),
        sa.Column("last_adx", sa.Float(), nullable=True),
        sa.Column("last_dmp", sa.Float(), nullable=True),
        sa.Column("last_dmn", sa.Float(), nullable=True),
        sa.Column("last_price", sa.Float(), nullable=True),
        sa.Column("last_price_date", sa.Date(), nullable=True),
        sa.Column("year_to_date_growth", sa.Float(), nullable=True),
        sa.Column("last_52_weeks_growth", sa.Float(), nullable=True),
        sa.Column("last_week_growth", sa.Float(), nullable=True),
        sa.Column("last_month_growth", sa.Float(), nullable=True),
        sa.Column("last_year_growth", sa.Float(), nullable=True),
        sa.Column("year_to_date_max_growth", sa.Float(), nullable=True),
        sa.Column("last_week_max_growth", sa.Float(), nullable=True),
        sa.Column("last_month_max_growth", sa.Float(), nullable=True),
        sa.Column("last_year_max_growth", sa.Float(), nullable=True),
        sa.Column("macd_12_26_9_buy", sa.Float(), nullable=True),
        sa.Column("star_yoy", sa.Float(), nullable=True),
        sa.Column("star_wow", sa.Float(), nullable=True),
        sa.Column("star_mom", sa.Float(), nullable=True),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("modifier", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("sector", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("industry_group", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("industry", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("website", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market_capitalization", sa.Float(), nullable=True),
        sa.Column("book_value", sa.Float(), nullable=True),
        sa.Column("price_to_book", sa.Float(), nullable=True),
        sa.Column("trailing_price_to_earnings", sa.Float(), nullable=True),
        sa.Column("dividend_yield", sa.Float(), nullable=True),
        sa.Column("dividend_rate", sa.Float(), nullable=True),
        sa.Column("trailing_earnings_per_share", sa.Float(), nullable=True),
        sa.Column("forward_earnings_per_share", sa.Float(), nullable=True),
        sa.Column("return_on_equity", sa.Float(), nullable=True),
        sa.Column("operating_margins", sa.Float(), nullable=True),
        sa.Column("gross_margins", sa.Float(), nullable=True),
        sa.Column("revenue_growth", sa.Float(), nullable=True),
        sa.Column("price_per_earning_ratio", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source"),
    )
    op.create_table(
        "view",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("view_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source"),
    )


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("view")
    op.drop_table("analysis")
    # ### end Alembic commands ###



import json
import logging
import os
from datetime import date
from typing import Optional, List, TYPE_CHECKING

from pydantic import BaseModel, Field
from openai import OpenAI

if TYPE_CHECKING:
    from bullish.database.crud import BullishDb

logger = logging.getLogger(__name__)


def prompt(ticker: str) -> str:
    return f"""
You are a financial analysis assistant.

Using the latest reliable public data from the web — including analyst price targets from multiple reputable 
sources — analyze the stock ticker {ticker}.

Return ONLY valid JSON matching EXACTLY the schema below — no explanations, no preamble, no markdown, no code 
fences, no extra text:

{{
    "high_price_target": float,  // Analyst consensus high price target in USD (based on multiple sources)
    "low_price_target": float,   // Analyst consensus low price target in USD (based on multiple sources)
    "recent_news": str,          // Detailed, multi-sentence summary of recent news affecting the company; 
    include credible source names inline
    "recommendation": str,       // One of: "Strong Buy", "Buy", "Hold", "Sell", "Strong Sell"
    "explanation": str           // Concise explanation for the recommendation above, covering key pros/cons 
    for investors
    "moat": bool           // Give as a boolean true or false if the company has a strong economic moat
}}

Formatting rules:
- Output must be a single valid JSON object with no surrounding text or formatting.
- Use plain numbers for high_price_target and low_price_target (no currency symbols, no commas).
- All text fields must be professional, investor-oriented, and reference credible named sources in `recent_news`.
- If exact data is unavailable, estimate based on web search results and note uncertainty in the relevant field.
"""


class OpenAINews(BaseModel):
    symbol: str
    news_date: date = Field(default_factory=date.today)
    high_price_target: Optional[float] = None
    low_price_target: Optional[float] = None
    recent_news: Optional[str] = None
    recommendation: Optional[str] = None
    explanation: Optional[str] = None
    moat: Optional[bool] = None

    def valid(self) -> bool:
        return bool(
            self.model_dump(
                exclude_none=True,
                exclude_unset=True,
                exclude_defaults=True,
                exclude={"symbol"},
            )
        )

    @classmethod
    def from_ticker(cls, ticker: str) -> "OpenAINews":
        if "OPENAI_API_KEY" not in os.environ:
            return cls(symbol=ticker)
        print(f"Fetching OpenAI news for {ticker}...")
        client = OpenAI()
        resp = client.responses.create(
            model="gpt-4o", input=prompt(ticker), tools=[{"type": "web_search"}]  # type: ignore
        )
        try:
            return cls.model_validate(json.loads(resp.output_text) | {"symbol": ticker})
        except Exception as e:
            logger.error(f"Failed to parse OpenAI response for {ticker}: {e}")
            return cls(symbol=ticker)

    @classmethod
    def from_tickers(cls, tickers: List[str]) -> List["OpenAINews"]:
        return [cls.from_ticker(t) for t in tickers]


def get_open_ai_news(bullish_db: "BullishDb", tickers: List[str]) -> bool:
    news = OpenAINews.from_tickers(tickers)
    valid_news = [n for n in news if n.valid()]
    if valid_news:
        bullish_db.write_many_openai_news(valid_news)
        return True
    return False



"""

Revision ID: ec25c8fa449f
Revises: 5b10ee7604c1
Create Date: 2025-07-14 12:26:29.346903

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "ec25c8fa449f"
down_revision: Union[str, None] = "5b10ee7604c1"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "industryreturns",
        sa.Column("date", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("created_at", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("simple_return", sa.Float(), nullable=True),
        sa.Column("log_return", sa.Float(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("industry", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("industry_group", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("sector", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("type", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.PrimaryKeyConstraint("date", "country", "industry", "type"),
    )

    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_analysis_return_after_rsi_crossover_45_period_90")
        )
        batch_op.drop_column("return_after_rsi_crossover_45_period_90")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "return_after_rsi_crossover_45_period_90", sa.FLOAT(), nullable=True
            )
        )
        batch_op.create_index(
            batch_op.f("ix_analysis_return_after_rsi_crossover_45_period_90"),
            ["return_after_rsi_crossover_45_period_90"],
            unique=False,
        )

    op.drop_table("industryreturns")
    # ### end Alembic commands ###



import sqlite3
from contextlib import contextmanager
from pathlib import Path
from sqlite3 import Connection
from typing import Generator, List

from bearish.database.schemas import *  # type: ignore # noqa: F403

from bullish.database.schemas import *  # noqa: F403


@contextmanager
def get_sqlite_connection(database_path: Path) -> Generator[Connection, None, None]:
    conn = sqlite3.connect(database_path)
    try:
        yield conn
    finally:
        conn.close()


def get_table_names(module_name: str) -> List[str]:
    return [
        value.__tablename__
        for _, value in locals().items()
        if issubclass(value, SQLModel)  # type: ignore # noqa: F405
        and hasattr(value, "__tablename__")
        and hasattr(value, "__table__")
        and module_name in value.__module__
    ]


def get_table_names_from_path(database_path: Path) -> List[str]:
    with get_sqlite_connection(database_path) as conn:
        cursor = conn.cursor()
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name NOT LIKE 'sqlite_%';"
        )
        tables = [row[0] for row in cursor.fetchall()]
        cursor.close()
        return tables


def empty_analysis_table(database_path: Path) -> bool:
    if "analysis" not in get_table_names_from_path(database_path):
        return True
    with get_sqlite_connection(database_path) as conn:
        cursor = conn.cursor()

        cursor.execute("SELECT COUNT(*) FROM analysis")
        count = cursor.fetchone()[0]
        conn.close()
        return bool(count == 0)


def _compatible_table(database_path: Path, module_name: str) -> bool:
    if not database_path.exists() and not database_path.is_file():
        raise FileNotFoundError(f"Database file {database_path} does not exist.")
    table_names = get_table_names(module_name)
    return set(table_names).issubset(get_table_names_from_path(database_path))


def compatible_bearish_database(database_path: Path) -> bool:
    return _compatible_table(database_path, "bearish.database.schemas")


def compatible_bullish_database(database_path: Path) -> bool:
    return _compatible_table(database_path, "bullish.database.schemas")



"""

Revision ID: 17e51420e7ad
Revises: d663166c531d
Create Date: 2025-07-10 17:35:02.376675

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = "17e51420e7ad"
down_revision: Union[str, None] = "d663166c531d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("rsi_bullish_crossover_30", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("rsi_bullish_crossover_40", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("rsi_bullish_crossover_45", sa.Date(), nullable=True)
        )
        batch_op.add_column(
            sa.Column(
                "return_after_rsi_crossover_45_period_90", sa.Float(), nullable=True
            )
        )
        batch_op.drop_index(batch_op.f("ix_analysis_rsi_bullish_crossover"))
        batch_op.create_index(
            "ix_analysis_return_after_rsi_crossover_45_period_90",
            ["return_after_rsi_crossover_45_period_90"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_rsi_bullish_crossover_30",
            ["rsi_bullish_crossover_30"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_rsi_bullish_crossover_40",
            ["rsi_bullish_crossover_40"],
            unique=False,
        )
        batch_op.create_index(
            "ix_analysis_rsi_bullish_crossover_45",
            ["rsi_bullish_crossover_45"],
            unique=False,
        )
        batch_op.drop_column("rsi_bullish_crossover")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("rsi_bullish_crossover", sa.DATE(), nullable=True)
        )
        batch_op.drop_index("ix_analysis_rsi_bullish_crossover_45")
        batch_op.drop_index("ix_analysis_rsi_bullish_crossover_40")
        batch_op.drop_index("ix_analysis_rsi_bullish_crossover_30")
        batch_op.drop_index("ix_analysis_return_after_rsi_crossover_45_period_90")
        batch_op.create_index(
            batch_op.f("ix_analysis_rsi_bullish_crossover"),
            ["rsi_bullish_crossover"],
            unique=False,
        )
        batch_op.drop_column("return_after_rsi_crossover_45_period_90")
        batch_op.drop_column("rsi_bullish_crossover_45")
        batch_op.drop_column("rsi_bullish_crossover_40")
        batch_op.drop_column("rsi_bullish_crossover_30")

    # ### end Alembic commands ###
